# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test data for MMM proto generator."""

from collections.abc import Sequence
import datetime

from mmm.v1 import mmm_pb2 as mmm_pb
from mmm.v1.common import date_interval_pb2 as date_interval_pb
from mmm.v1.fit import model_fit_pb2 as fit_pb
from mmm.v1.marketing.analysis import marketing_analysis_pb2
from mmm.v1.marketing.optimization import budget_optimization_pb2 as budget_pb
from mmm.v1.marketing.optimization import reach_frequency_optimization_pb2 as rf_pb
from schema.processors import budget_optimization_processor
from schema.processors import marketing_processor
from schema.processors import model_fit_processor
from schema.processors import model_processor
from schema.processors import reach_frequency_optimization_processor as rf_opt_processor

from google.type import date_pb2

# Weekly dates from 2022-11-21 to 2024-01-01.
ALL_TIMES_IN_MERIDIAN = (
    '2022-11-21',
    '2022-11-28',
    '2022-12-05',
    '2022-12-12',
    '2022-12-19',
    '2022-12-26',
    '2023-01-02',
    '2023-01-09',
    '2023-01-16',
    '2023-01-23',
    '2023-01-30',
    '2023-02-06',
    '2023-02-13',
    '2023-02-20',
    '2023-02-27',
    '2023-03-06',
    '2023-03-13',
    '2023-03-20',
    '2023-03-27',
    '2023-04-03',
    '2023-04-10',
    '2023-04-17',
    '2023-04-24',
    '2023-05-01',
    '2023-05-08',
    '2023-05-15',
    '2023-05-22',
    '2023-05-29',
    '2023-06-05',
    '2023-06-12',
    '2023-06-19',
    '2023-06-26',
    '2023-07-03',
    '2023-07-10',
    '2023-07-17',
    '2023-07-24',
    '2023-07-31',
    '2023-08-07',
    '2023-08-14',
    '2023-08-21',
    '2023-08-28',
    '2023-09-04',
    '2023-09-11',
    '2023-09-18',
    '2023-09-25',
    '2023-10-02',
    '2023-10-09',
    '2023-10-16',
    '2023-10-23',
    '2023-10-30',
    '2023-11-06',
    '2023-11-13',
    '2023-11-20',
    '2023-11-27',
    '2023-12-04',
    '2023-12-11',
    '2023-12-18',
    '2023-12-25',
    '2024-01-01',
)

ALL_TIME_BUCKET_DATED_SPECS = (
    # All
    model_processor.DatedSpec(
        start_date=datetime.date(2022, 11, 21),
        end_date=datetime.date(2024, 1, 8),
        date_interval_tag='ALL',
    ),
    # Monthly buckets
    model_processor.DatedSpec(
        start_date=datetime.date(2022, 12, 5),
        end_date=datetime.date(2023, 1, 2),
        date_interval_tag='Y2022 Dec',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 1, 2),
        end_date=datetime.date(2023, 2, 6),
        date_interval_tag='Y2023 Jan',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 2, 6),
        end_date=datetime.date(2023, 3, 6),
        date_interval_tag='Y2023 Feb',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 3, 6),
        end_date=datetime.date(2023, 4, 3),
        date_interval_tag='Y2023 Mar',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 4, 3),
        end_date=datetime.date(2023, 5, 1),
        date_interval_tag='Y2023 Apr',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 5, 1),
        end_date=datetime.date(2023, 6, 5),
        date_interval_tag='Y2023 May',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 6, 5),
        end_date=datetime.date(2023, 7, 3),
        date_interval_tag='Y2023 Jun',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 7, 3),
        end_date=datetime.date(2023, 8, 7),
        date_interval_tag='Y2023 Jul',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 8, 7),
        end_date=datetime.date(2023, 9, 4),
        date_interval_tag='Y2023 Aug',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 9, 4),
        end_date=datetime.date(2023, 10, 2),
        date_interval_tag='Y2023 Sep',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 10, 2),
        end_date=datetime.date(2023, 11, 6),
        date_interval_tag='Y2023 Oct',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 11, 6),
        end_date=datetime.date(2023, 12, 4),
        date_interval_tag='Y2023 Nov',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 12, 4),
        end_date=datetime.date(2024, 1, 1),
        date_interval_tag='Y2023 Dec',
    ),
    # Quarterly buckets
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 1, 2),
        end_date=datetime.date(2023, 4, 3),
        date_interval_tag='Y2023 Q1',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 4, 3),
        end_date=datetime.date(2023, 7, 3),
        date_interval_tag='Y2023 Q2',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 7, 3),
        end_date=datetime.date(2023, 10, 2),
        date_interval_tag='Y2023 Q3',
    ),
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 10, 2),
        end_date=datetime.date(2024, 1, 1),
        date_interval_tag='Y2023 Q4',
    ),
    # Yearly buckets
    model_processor.DatedSpec(
        start_date=datetime.date(2023, 1, 2),
        end_date=datetime.date(2024, 1, 1),
        date_interval_tag='Y2023',
    ),
)


def _dated_spec_to_date_interval(
    spec: model_processor.DatedSpec,
) -> date_interval_pb.DateInterval:
  if spec.start_date is None or spec.end_date is None:
    raise ValueError('Start date or end date is None.')

  return date_interval_pb.DateInterval(
      start_date=date_pb2.Date(
          year=spec.start_date.year,
          month=spec.start_date.month,
          day=spec.start_date.day,
      ),
      end_date=date_pb2.Date(
          year=spec.end_date.year,
          month=spec.end_date.month,
          day=spec.end_date.day,
      ),
      tag=spec.date_interval_tag,
  )


class FakeModelFitProcessor(
    model_processor.ModelProcessor[
        model_fit_processor.ModelFitSpec, fit_pb.ModelFit
    ]
):
  """Fake ModelFitProcessor for testing."""

  def __init__(self, trained_model: model_processor.TrainedModel):
    self._trained_model = trained_model

  @classmethod
  def spec_type(cls):
    return model_fit_processor.ModelFitSpec

  @classmethod
  def output_type(cls):
    return fit_pb.ModelFit

  def execute(
      self, specs: Sequence[model_fit_processor.ModelFitSpec]
  ) -> fit_pb.ModelFit:
    return fit_pb.ModelFit()

  def _set_output(self, output: mmm_pb.Mmm, result: fit_pb.ModelFit):
    output.model_fit.CopyFrom(result)


class FakeBudgetOptimizationProcessor(
    model_processor.ModelProcessor[
        budget_optimization_processor.BudgetOptimizationSpec,
        budget_pb.BudgetOptimization,
    ]
):
  """Fake BudgetOptimizationProcessor for testing."""

  def __init__(self, trained_model: model_processor.TrainedModel):
    self._trained_model = trained_model

  @classmethod
  def spec_type(cls):
    return budget_optimization_processor.BudgetOptimizationSpec

  @classmethod
  def output_type(cls):
    return budget_pb.BudgetOptimization

  def execute(
      self,
      specs: Sequence[budget_optimization_processor.BudgetOptimizationSpec],
  ) -> budget_pb.BudgetOptimization:
    results = []
    for spec in specs:
      result = budget_pb.BudgetOptimizationResult(
          name=spec.optimization_name,
          spec=budget_pb.BudgetOptimizationSpec(
              date_interval=_dated_spec_to_date_interval(spec)
          ),
          incremental_outcome_grid=budget_pb.IncrementalOutcomeGrid(
              name=spec.grid_name
          ),
      )
      if spec.group_id:
        result.group_id = spec.group_id
      results.append(result)

    return budget_pb.BudgetOptimization(results=results)

  def _set_output(
      self, output: mmm_pb.Mmm, result: budget_pb.BudgetOptimization
  ):
    output.marketing_optimization.budget_optimization.CopyFrom(result)


class FakeReachFrequencyOptimizationProcessor(
    model_processor.ModelProcessor[
        rf_opt_processor.ReachFrequencyOptimizationSpec,
        rf_pb.ReachFrequencyOptimization,
    ]
):
  """Fake ReachFrequencyOptimizationProcessor for testing."""

  def __init__(self, trained_model: model_processor.TrainedModel):
    self._trained_model = trained_model

  @classmethod
  def spec_type(cls):
    return rf_opt_processor.ReachFrequencyOptimizationSpec

  @classmethod
  def output_type(cls):
    return rf_pb.ReachFrequencyOptimization

  def execute(
      self,
      specs: Sequence[rf_opt_processor.ReachFrequencyOptimizationSpec],
  ) -> rf_pb.ReachFrequencyOptimization:
    results = []
    for spec in specs:
      result = rf_pb.ReachFrequencyOptimizationResult(
          name=spec.optimization_name,
          spec=rf_pb.ReachFrequencyOptimizationSpec(
              date_interval=_dated_spec_to_date_interval(spec)
          ),
          frequency_outcome_grid=rf_pb.FrequencyOutcomeGrid(
              name=spec.grid_name
          ),
      )
      if spec.group_id:
        result.group_id = spec.group_id
      results.append(result)

    return rf_pb.ReachFrequencyOptimization(results=results)

  def _set_output(
      self,
      output: mmm_pb.Mmm,
      result: rf_pb.ReachFrequencyOptimization,
  ):
    output.marketing_optimization.reach_frequency_optimization.CopyFrom(result)


class FakeMarketingProcessor(
    model_processor.ModelProcessor[
        marketing_processor.MarketingAnalysisSpec,
        marketing_analysis_pb2.MarketingAnalysisList,
    ]
):
  """Fake MarketingProcessor for testing."""

  def __init__(self, trained_model: model_processor.TrainedModel):
    self._trained_model = trained_model

  @classmethod
  def spec_type(cls):
    return marketing_processor.MarketingAnalysisSpec

  @classmethod
  def output_type(cls):
    return marketing_analysis_pb2.MarketingAnalysisList

  def execute(
      self, specs: Sequence[marketing_processor.MarketingAnalysisSpec]
  ) -> marketing_analysis_pb2.MarketingAnalysisList:
    marketing_analyses = []
    for spec in specs:
      marketing_analysis = marketing_analysis_pb2.MarketingAnalysis(
          date_interval=_dated_spec_to_date_interval(spec)
      )
      marketing_analyses.append(marketing_analysis)

    return marketing_analysis_pb2.MarketingAnalysisList(
        marketing_analyses=marketing_analyses
    )

  def _set_output(
      self,
      output: mmm_pb.Mmm,
      result: marketing_analysis_pb2.MarketingAnalysisList,
  ):
    output.marketing_analysis_list.CopyFrom(result)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An output converter that denormalizes into flat data frame tables."""

from collections.abc import Mapping, Sequence

from scenarioplanner.converters import mmm_converter
from scenarioplanner.converters.dataframe import budget_opt_converters
from scenarioplanner.converters.dataframe import converter
from scenarioplanner.converters.dataframe import marketing_analyses_converters
from scenarioplanner.converters.dataframe import rf_opt_converters
import pandas as pd


__all__ = ["DataFrameModelConverter"]


class DataFrameModelConverter(mmm_converter.ModelConverter[pd.DataFrame]):
  """Converts a bound `Mmm` model into denormalized flat data frame tables.

  The denormalized, two-dimensional data frame tables are intended to be
  directly compiled into sheets in a Google Sheets file to be used as a data
  source for a Looker Studio dashboard.

  These data frame tables are:

  *   "ModelDiagnostics"
  *   "ModelFit"
  *   "MediaOutcome"
  *   "MediaSpend"
  *   "MediaROI"
  *   (Named Incremental Outcome Grids)
  *   "budget_opt_specs"
  *   "budget_opt_results"
  *   "response_curves"
  *   (Named R&F ROI Grids)
  *   "rf_opt_specs"
  *   "rf_opt_results"
  """

  _converters: Sequence[type[converter.Converter]] = (
      marketing_analyses_converters.CONVERTERS
      + budget_opt_converters.CONVERTERS
      + rf_opt_converters.CONVERTERS
  )

  def __call__(self, **kwargs) -> Mapping[str, pd.DataFrame]:
    """Converts bound `Mmm` model proto to named, flat data frame tables."""
    output = {}

    for converter_class in self._converters:
      converter_instance = converter_class(self.mmm)  # pytype: disable=not-instantiable
      for table_name, table_data in converter_instance():
        if output.get(table_name) is not None:
          raise ValueError(f"Duplicate table name: {table_name}")
        output[table_name] = table_data

    return output



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Function registry for Serde."""

from __future__ import annotations

import functools
import hashlib
import inspect
from typing import Any, Callable
import warnings


class SourceCodeRetrievalError(Exception):
  """Raised when the source code of a function cannot be retrieved."""


class LambdaSourceCodeWarning(UserWarning):
  """Warning issued when trying to get source code of a lambda function."""


def _get_func_source(func: Callable[..., Any]) -> str:
  """Returns the source code of a function.

  Args:
    func: The function to get the source code for.

  Returns:
    The source code of the function.

  Raises:
    SourceCodeRetrievalError: If the source code of the function cannot be
      retrieved.
  """
  if hasattr(func, "__code__") and func.__code__.co_name == "<lambda>":
    warnings.warn(
        "Retrieving the source code of a lambda function might not work"
        " successfully. It's recommended to explicitly define a function.",
        LambdaSourceCodeWarning,
    )
  try:
    return inspect.getsource(func)
  except (OSError, TypeError) as e:
    raise SourceCodeRetrievalError(
        f"Source code of function {func} is not retrievable."
    ) from e


def _get_hash(value: str) -> str:
  """Returns a SHA-256 hash of the given value."""
  encoded_string = value.encode("utf-8")
  sha_256_hash = hashlib.sha256()
  sha_256_hash.update(encoded_string)
  return sha_256_hash.hexdigest()


class FunctionRegistry(dict[str, Callable[..., Any]]):
  """A dictionary-like container for custom functions used in serialization.

  This class extends dict and provides methods for hashing, validation,
  and key retrieval based on function identity, required for safe
  serialization and deserialization of models that use custom functions.
  """

  def __init__(self, *args, **kwargs):
    """Initializes the FunctionRegistry.

    Accepts the same arguments as a standard dictionary constructor.
    For example:
      reg = FunctionRegistry({'func1': my_func1, 'func2': my_func2})
      reg = FunctionRegistry(func1=my_func1, func2=my_func2)

    Args:
      *args: Positional arguments to pass to the dictionary constructor.
      **kwargs: Keyword arguments to pass to the dictionary constructor.
    """
    super().__init__(*args, **kwargs)

  @functools.cached_property
  def hashed_registry(self) -> dict[str, str]:
    """Returns hashed function registry with keys mapped to hashed function code."""
    return {
        key: _get_hash(_get_func_source(function))
        for key, function in self.items()
    }

  def validate(self, stored_hashed_function_registry: dict[str, str]):
    """Validates whether functions within the registry have changed.

    It checks that all functions listed in stored_hashed_function_registry
    are present in this registry, and that their source code hash matches
    the stored hash.

    Args:
      stored_hashed_function_registry: The hashed function registry from the
        serialized object.

    Raises:
      ValueError: If a function is missing or a hash mismatch is detected.
    """
    if not stored_hashed_function_registry and self:
      warnings.warn(
          "A function registry was provided during loading, but none was"
          " found on the serialized object. Custom functions will be"
          " ignored."
      )
      return

    for key, stored_hash in stored_hashed_function_registry.items():
      if key not in self:
        raise ValueError(
            f"Function '{key}' is required by the serialized object but"
            " is missing from the provided function registry."
        )
      func = self[key]
      try:
        source_code = _get_func_source(func)
      except SourceCodeRetrievalError as e:
        raise ValueError(
            f"Failed to retrieve source code of function {key}."
        ) from e
      evaluated_hash = _get_hash(source_code)
      if stored_hash != evaluated_hash:
        raise ValueError(f"Function registry hash mismatch for {key}.")

  def get_function_key(self, func: Callable[..., Any]) -> str | None:
    """Returns the function key for the given function from the registry."""
    for function_key, registry_func in self.items():
      if func is registry_func:
        return function_key
    return None



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared test data."""

from collections.abc import Iterator, Sequence

from meridian import constants as c
from mmm.v1.common import date_interval_pb2 as date_interval_pb
from mmm.v1.common import estimate_pb2 as estimate_pb
from mmm.v1.common import kpi_type_pb2 as kpi_type_pb
from mmm.v1.common import target_metric_pb2 as target_metric_pb
from mmm.v1.fit import model_fit_pb2 as fit_pb
from mmm.v1.marketing import marketing_data_pb2 as marketing_data_pb
from mmm.v1.marketing.analysis import marketing_analysis_pb2 as marketing_pb
from mmm.v1.marketing.analysis import media_analysis_pb2 as media_pb
from mmm.v1.marketing.analysis import non_media_analysis_pb2 as non_media_pb
from mmm.v1.marketing.analysis import outcome_pb2 as outcome_pb
from mmm.v1.marketing.analysis import response_curve_pb2 as response_curve_pb
from mmm.v1.marketing.optimization import budget_optimization_pb2 as budget_pb
from mmm.v1.marketing.optimization import constraints_pb2 as constraints_pb
from mmm.v1.marketing.optimization import reach_frequency_optimization_pb2 as rf_pb
from scenarioplanner.converters.dataframe import constants as cc

from google.type import date_pb2 as date_pb


DATES = [
    date_pb.Date(year=2024, month=1, day=1),
    date_pb.Date(year=2024, month=1, day=8),
    date_pb.Date(year=2024, month=1, day=15),
]
DATE_INTERVALS = [
    date_interval_pb.DateInterval(
        start_date=DATES[0],
        end_date=DATES[1],
        tag="Week1",
    ),
    date_interval_pb.DateInterval(
        start_date=DATES[1],
        end_date=DATES[2],
        tag="Week2",
    ),
]
ALL_DATE_INTERVAL = date_interval_pb.DateInterval(
    start_date=DATES[0],
    end_date=DATES[2],
    tag=cc.ANALYSIS_TAG_ALL,
)


GEO_INFOS = [
    marketing_data_pb.GeoInfo(
        geo_id="geo-1",
        population=100,
    ),
    marketing_data_pb.GeoInfo(
        geo_id="geo-2",
        population=200,
    ),
]


MEDIA_CHANNELS = [
    "Channel 1",
    "Channel 2",
]
RF_CHANNELS = [
    "RF Channel 1",
    "RF Channel 2",
]


BASE_MEDIA_SPEND = 100.0
BASE_RF_MEDIA_SPEND = 110.0


def _create_marketing_data(
    create_rf_data: bool = True,
) -> Iterator[marketing_data_pb.MarketingDataPoint]:
  """Generator for default `MarketingDataPoint`s for each geo and date interval defined above."""
  for geo_info in GEO_INFOS:
    for date_interval in DATE_INTERVALS:
      media_vars = []
      rf_vars = []
      for channel in MEDIA_CHANNELS:
        media_var = marketing_data_pb.MediaVariable(
            channel_name=channel,
            # For simplicity, set all media spend to be the same across all
            # channels and across all geo and time dimensions.
            # Add function parameters if more sophisticated test data
            # generator is warranted here.
            media_spend=BASE_MEDIA_SPEND,
        )
        media_vars.append(media_var)
      if create_rf_data:
        for channel in RF_CHANNELS:
          rf_media_var = marketing_data_pb.ReachFrequencyVariable(
              channel_name=channel,
              spend=BASE_RF_MEDIA_SPEND,
              reach=10_000,
              average_frequency=1.1,
          )
          rf_vars.append(rf_media_var)
      yield marketing_data_pb.MarketingDataPoint(
          date_interval=date_interval,
          geo_info=geo_info,
          media_variables=media_vars,
          reach_frequency_variables=rf_vars,
          # `kpi` and `control_variables` fields are not set, since no test
          # needs it just yet. Fill them in when needed.
      )


MARKETING_DATA = marketing_data_pb.MarketingData(
    marketing_data_points=list(_create_marketing_data()),
)


PERFORMANCE_TEST = fit_pb.Performance(
    r_squared=0.99,
    mape=67.7,
    weighted_mape=59.8,
    rmse=55.05,
)
PERFORMANCE_TRAIN = fit_pb.Performance(
    r_squared=0.91,
    mape=60.6,
    weighted_mape=55.5,
    rmse=59.87,
)
PERFORMANCE_ALL_DATA = fit_pb.Performance(
    r_squared=0.94,
    mape=60.0,
    weighted_mape=55.4,
    rmse=52.0,
)


def _create_model_fit_result(
    name: str,
    performance: fit_pb.Performance,
) -> fit_pb.Result:
  return fit_pb.Result(
      name=name,
      performance=performance,
      predictions=[
          fit_pb.Prediction(
              date_interval=DATE_INTERVALS[0],
              predicted_outcome=estimate_pb.Estimate(
                  value=100.0,
                  uncertainties=[
                      estimate_pb.Estimate.Uncertainty(
                          probability=0.9,
                          lowerbound=90.0,
                          upperbound=110.0,
                      )
                  ],
              ),
              predicted_baseline=estimate_pb.Estimate(
                  value=90.0,
                  uncertainties=[
                      estimate_pb.Estimate.Uncertainty(
                          probability=0.9,
                          lowerbound=89.0,
                          upperbound=111.0,
                      )
                  ],
              ),
              actual_value=105.0,
          ),
          fit_pb.Prediction(
              date_interval=DATE_INTERVALS[1],
              predicted_outcome=estimate_pb.Estimate(
                  value=110.0,
                  uncertainties=[
                      estimate_pb.Estimate.Uncertainty(
                          probability=0.9,
                          lowerbound=100.0,
                          upperbound=120.0,
                      )
                  ],
              ),
              predicted_baseline=estimate_pb.Estimate(
                  value=109.0,
                  uncertainties=[
                      estimate_pb.Estimate.Uncertainty(
                          probability=0.9,
                          lowerbound=90.0,
                          upperbound=125.0,
                      )
                  ],
              ),
              actual_value=115.0,
          ),
      ],
  )


MODEL_FIT_RESULT_TEST = _create_model_fit_result(
    name=c.TEST,
    performance=PERFORMANCE_TEST,
)
MODEL_FIT_RESULT_TRAIN = _create_model_fit_result(
    name=c.TRAIN,
    performance=PERFORMANCE_TRAIN,
)
MODEL_FIT_RESULT_ALL_DATA = _create_model_fit_result(
    name=c.ALL_DATA,
    performance=PERFORMANCE_ALL_DATA,
)


def create_outcome(
    incremental_outcome: float,
    pct_of_contribution: float,
    effectiveness: float,
    roi: float,
    mroi: float,
    cpik: float,
    is_revenue_type: bool,
) -> outcome_pb.Outcome:
  return outcome_pb.Outcome(
      kpi_type=(
          kpi_type_pb.REVENUE if is_revenue_type else kpi_type_pb.NON_REVENUE
      ),
      contribution=outcome_pb.Contribution(
          value=estimate_pb.Estimate(value=incremental_outcome),
          share=estimate_pb.Estimate(value=pct_of_contribution),
      ),
      effectiveness=outcome_pb.Effectiveness(
          media_unit=c.IMPRESSIONS,
          value=estimate_pb.Estimate(value=effectiveness),
      ),
      roi=estimate_pb.Estimate(
          value=roi,
          uncertainties=[
              estimate_pb.Estimate.Uncertainty(
                  probability=0.9,
                  lowerbound=roi * 0.9,
                  upperbound=roi * 1.1,
              )
          ],
      ),
      marginal_roi=estimate_pb.Estimate(value=mroi),
      cost_per_contribution=estimate_pb.Estimate(
          value=cpik,
          uncertainties=[
              estimate_pb.Estimate.Uncertainty(
                  probability=0.9,
                  lowerbound=cpik * 0.9,
                  upperbound=cpik * 1.1,
              )
          ],
      ),
  )


REVENUE_OUTCOME = create_outcome(
    incremental_outcome=100.0,
    pct_of_contribution=0.1,
    effectiveness=3.3,
    roi=1.0,
    mroi=10.0,
    cpik=5.0,
    is_revenue_type=True,
)

NON_REVENUE_OUTCOME = create_outcome(
    incremental_outcome=100.0,
    pct_of_contribution=0.1,
    effectiveness=4.4,
    roi=10.0,
    mroi=100.0,
    cpik=100.0,
    is_revenue_type=False,
)


SPENDS = {
    MEDIA_CHANNELS[0]: 75_000,
    MEDIA_CHANNELS[1]: 25_000,
    RF_CHANNELS[0]: 30_000,
    RF_CHANNELS[1]: 20_000,
}
TOTAL_SPEND = sum(SPENDS.values())
SPENDS[c.ALL_CHANNELS] = TOTAL_SPEND


def create_media_analysis(
    channel: str,
    multiplier: float = 1.0,
    make_revenue_outcome: bool = True,
    make_non_revenue_outcome: bool = True,
) -> media_pb.MediaAnalysis:
  """Creates a `MediaAnalysis` test proto."""
  # `multiplier` is used to create unique metric numbers for the given channel
  # from the base template metrics above.
  outcomes = []
  if make_revenue_outcome:
    outcomes.append(
        create_outcome(
            incremental_outcome=100.0 * multiplier,
            pct_of_contribution=0.1 * multiplier,
            effectiveness=2.2 * multiplier,
            roi=1.0 * multiplier,
            mroi=10.0 * multiplier,
            cpik=5.0 * multiplier,
            is_revenue_type=True,
        )
    )
  if make_non_revenue_outcome:
    outcomes.append(
        create_outcome(
            incremental_outcome=100.0 * multiplier,
            pct_of_contribution=0.1 * multiplier,
            effectiveness=5.5 * multiplier,
            roi=10.0 * multiplier,
            mroi=100.0 * multiplier,
            cpik=100.0 * multiplier,
            is_revenue_type=False,
        )
    )

  response_curve = response_curve_pb.ResponseCurve(
      input_name="Spend",
      response_points=[
          response_curve_pb.ResponsePoint(
              input_value=1 * multiplier,
              incremental_kpi=100.0 * multiplier,
          ),
          response_curve_pb.ResponsePoint(
              input_value=2 * multiplier,
              incremental_kpi=200.0 * multiplier,
          ),
      ],
  )
  return media_pb.MediaAnalysis(
      channel_name=channel,
      spend_info=media_pb.SpendInfo(
          spend=SPENDS[channel],
          spend_share=SPENDS[channel] / TOTAL_SPEND,
      ),
      media_outcomes=outcomes,
      response_curve=response_curve,
  )


MEDIA_ANALYSES_BOTH_OUTCOMES = [
    create_media_analysis(channel, (idx + 1))
    for (idx, channel) in enumerate(MEDIA_CHANNELS)
]
RF_ANALYSES_BOTH_OUTCOMES = [
    create_media_analysis(channel, (idx + 1))
    for (idx, channel) in enumerate(RF_CHANNELS)
]
MEDIA_ANALYSES_NONREVENUE = [
    create_media_analysis(
        channel,
        # use a different multiplier value to distinquish from the above
        (idx + 1.2),
        make_revenue_outcome=False,
    )
    for (idx, channel) in enumerate(MEDIA_CHANNELS)
]
RF_ANALYSES_NONREVENUE = [
    create_media_analysis(
        channel,
        # use a different multiplier value to distinquish from the above
        (idx + 1.2),
        make_revenue_outcome=False,
    )
    for (idx, channel) in enumerate(RF_CHANNELS)
]

ALL_CHANNELS_ANALYSIS_BOTH_OUTCOMES = create_media_analysis(
    c.ALL_CHANNELS, multiplier=10
)
ALL_CHANNELS_ANALYSIS_NONREVENUE = create_media_analysis(
    c.ALL_CHANNELS, multiplier=12, make_revenue_outcome=False
)

BASELINE_NONREVENUE_OUTCOME = create_outcome(
    incremental_outcome=40.0,
    pct_of_contribution=0.04,
    effectiveness=4.4,
    cpik=75.0,
    roi=7.0,
    mroi=70.0,
    is_revenue_type=False,
)
BASELINE_ANALYSIS_NONREVENUE = non_media_pb.NonMediaAnalysis(
    non_media_name=c.BASELINE,
    non_media_outcomes=[BASELINE_NONREVENUE_OUTCOME],
)

BASELINE_REVENUE_OUTCOME = create_outcome(
    incremental_outcome=50.0,
    pct_of_contribution=0.05,
    effectiveness=5.5,
    roi=1.0,
    mroi=10.0,
    cpik=0.5,
    is_revenue_type=True,
)
BASELINE_ANALYSIS_REVENUE = non_media_pb.NonMediaAnalysis(
    non_media_name=c.BASELINE,
    non_media_outcomes=[BASELINE_REVENUE_OUTCOME],
)

BASELINE_ANALYSIS_BOTH_OUTCOMES = non_media_pb.NonMediaAnalysis(
    non_media_name=c.BASELINE,
    non_media_outcomes=[BASELINE_NONREVENUE_OUTCOME, BASELINE_REVENUE_OUTCOME],
)


def create_marketing_analysis(
    date_interval: date_interval_pb.DateInterval,
    baseline_analysis: non_media_pb.NonMediaAnalysis = BASELINE_ANALYSIS_BOTH_OUTCOMES,
    explicit_channel_analyses: Sequence[media_pb.MediaAnalysis] | None = None,
    explicit_all_channels_analysis: media_pb.MediaAnalysis | None = None,
) -> marketing_pb.MarketingAnalysis:
  """Create a `MarketingAnalysis` for the given analysis period and tag."""
  media_analyses = (
      list(explicit_channel_analyses)
      if explicit_channel_analyses
      else (MEDIA_ANALYSES_BOTH_OUTCOMES + RF_ANALYSES_BOTH_OUTCOMES)
  )
  media_analyses.append(
      explicit_all_channels_analysis
      if explicit_all_channels_analysis
      else ALL_CHANNELS_ANALYSIS_BOTH_OUTCOMES
  )

  return marketing_pb.MarketingAnalysis(
      date_interval=date_interval,
      non_media_analyses=[baseline_analysis],
      media_analyses=media_analyses,
  )


# All of the below test analyses data contain both media and R&F channels.

ALL_TAG_MARKETING_ANALYSIS_BOTH_OUTCOMES = create_marketing_analysis(
    date_interval=ALL_DATE_INTERVAL,
    baseline_analysis=BASELINE_ANALYSIS_BOTH_OUTCOMES,
)
ALL_TAG_MARKETING_ANALYSIS_NONREVENUE = create_marketing_analysis(
    date_interval=ALL_DATE_INTERVAL,
    baseline_analysis=BASELINE_ANALYSIS_NONREVENUE,
    explicit_channel_analyses=(
        MEDIA_ANALYSES_NONREVENUE + RF_ANALYSES_NONREVENUE
    ),
)

DATED_MARKETING_ANALYSES_BOTH_OUTCOMES = [
    create_marketing_analysis(
        date_interval=date_interval,
        baseline_analysis=BASELINE_ANALYSIS_BOTH_OUTCOMES,
    )
    for date_interval in DATE_INTERVALS
]
DATED_MARKETING_ANALYSES_NONREVENUE = [
    create_marketing_analysis(
        date_interval=date_interval,
        baseline_analysis=BASELINE_ANALYSIS_NONREVENUE,
        explicit_channel_analyses=(
            MEDIA_ANALYSES_NONREVENUE + RF_ANALYSES_NONREVENUE
        ),
    )
    for date_interval in DATE_INTERVALS
]

MARKETING_ANALYSIS_LIST_BOTH_OUTCOMES = marketing_pb.MarketingAnalysisList(
    marketing_analyses=(
        [ALL_TAG_MARKETING_ANALYSIS_BOTH_OUTCOMES]
        + DATED_MARKETING_ANALYSES_BOTH_OUTCOMES
    ),
)

MARKETING_ANALYSIS_LIST_NONREVENUE = marketing_pb.MarketingAnalysisList(
    marketing_analyses=(
        [ALL_TAG_MARKETING_ANALYSIS_NONREVENUE]
        + DATED_MARKETING_ANALYSES_NONREVENUE
    ),
)


# Incremental outcome grids (budget) are only relevant for non-RF media
# channels.

INCREMENTAL_OUTCOME_GRID_FOO = budget_pb.IncrementalOutcomeGrid(
    name="incremental outcome grid foo",
    channel_cells=[
        budget_pb.IncrementalOutcomeGrid.ChannelCells(
            channel_name=MEDIA_CHANNELS[0],
            cells=[
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=10000.0,
                    incremental_outcome=estimate_pb.Estimate(value=100.0),
                ),
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=20000.0,
                    incremental_outcome=estimate_pb.Estimate(value=200.0),
                ),
            ],
        ),
        budget_pb.IncrementalOutcomeGrid.ChannelCells(
            channel_name=MEDIA_CHANNELS[1],
            cells=[
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=10000.0,
                    incremental_outcome=estimate_pb.Estimate(value=100.0),
                ),
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=20000.0,
                    incremental_outcome=estimate_pb.Estimate(value=200.0),
                ),
            ],
        ),
    ],
)

INCREMENTAL_OUTCOME_GRID_BAR = budget_pb.IncrementalOutcomeGrid(
    name="incremental outcome grid bar",
    channel_cells=[
        budget_pb.IncrementalOutcomeGrid.ChannelCells(
            channel_name=MEDIA_CHANNELS[0],
            cells=[
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=1000.0,
                    incremental_outcome=estimate_pb.Estimate(value=10.0),
                ),
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=2000.0,
                    incremental_outcome=estimate_pb.Estimate(value=20.0),
                ),
            ],
        ),
        budget_pb.IncrementalOutcomeGrid.ChannelCells(
            channel_name=MEDIA_CHANNELS[1],
            cells=[
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=1000.0,
                    incremental_outcome=estimate_pb.Estimate(value=10.0),
                ),
                budget_pb.IncrementalOutcomeGrid.Cell(
                    spend=2000.0,
                    incremental_outcome=estimate_pb.Estimate(value=20.0),
                ),
            ],
        ),
    ],
)

# A fixed budget scenario for the entire time interval in the test data above.
BUDGET_OPTIMIZATION_SPEC_FIXED_ALL_DATES = budget_pb.BudgetOptimizationSpec(
    date_interval=ALL_DATE_INTERVAL,
    objective=target_metric_pb.TargetMetric.ROI,
    fixed_budget_scenario=budget_pb.FixedBudgetScenario(total_budget=100000.0),
    # No individual channel constraints. Expect implicit constraints: max budget
    # applied for each channel.
)
BUDGET_OPTIMIZATION_RESULT_FIXED_BOTH_OUTCOMES = (
    budget_pb.BudgetOptimizationResult(
        name="budget optimization result foo",
        group_id="group-foo",
        optimized_marketing_analysis=ALL_TAG_MARKETING_ANALYSIS_BOTH_OUTCOMES,
        spec=BUDGET_OPTIMIZATION_SPEC_FIXED_ALL_DATES,
        incremental_outcome_grid=INCREMENTAL_OUTCOME_GRID_FOO,
    )
)

# A flexible budget scenario for the second time interval only.
BUDGET_OPTIMIZATION_SPEC_FLEX_SELECT_DATES = budget_pb.BudgetOptimizationSpec(
    date_interval=DATE_INTERVALS[1],
    objective=target_metric_pb.TargetMetric.KPI,
    flexible_budget_scenario=budget_pb.FlexibleBudgetScenario(
        total_budget_constraint=constraints_pb.BudgetConstraint(
            min_budget=1000.0,
            max_budget=2000.0,
        ),
        target_metric_constraints=[
            constraints_pb.TargetMetricConstraint(
                target_metric=target_metric_pb.COST_PER_INCREMENTAL_KPI,
                target_value=10.0,
            )
        ],
    ),
    # Define explicit channel constraints.
    channel_constraints=[
        budget_pb.ChannelConstraint(
            channel_name=MEDIA_CHANNELS[0],
            budget_constraint=constraints_pb.BudgetConstraint(
                min_budget=1100.0,
                max_budget=1500.0,
            ),
        ),
        budget_pb.ChannelConstraint(
            channel_name=MEDIA_CHANNELS[1],
            budget_constraint=constraints_pb.BudgetConstraint(
                min_budget=1000.0,
                max_budget=1800.0,
            ),
        ),
    ],
)
BUDGET_OPTIMIZATION_RESULT_FLEX_NONREV = budget_pb.BudgetOptimizationResult(
    name="budget optimization result bar",
    group_id="group-bar",
    optimized_marketing_analysis=ALL_TAG_MARKETING_ANALYSIS_NONREVENUE,
    spec=BUDGET_OPTIMIZATION_SPEC_FLEX_SELECT_DATES,
    incremental_outcome_grid=INCREMENTAL_OUTCOME_GRID_BAR,
)


# Frequency outcome grids are only relevant for R&F media channels.

FREQUENCY_OUTCOME_GRID_FOO = rf_pb.FrequencyOutcomeGrid(
    name="frequency outcome grid foo",
    channel_cells=[
        rf_pb.FrequencyOutcomeGrid.ChannelCells(
            channel_name=RF_CHANNELS[0],
            cells=[
                rf_pb.FrequencyOutcomeGrid.Cell(
                    reach_frequency=rf_pb.ReachFrequency(
                        reach=10000,
                        average_frequency=1.0,
                    ),
                    outcome=estimate_pb.Estimate(value=100.0),
                ),
                rf_pb.FrequencyOutcomeGrid.Cell(
                    reach_frequency=rf_pb.ReachFrequency(
                        reach=20000,
                        average_frequency=2.0,
                    ),
                    outcome=estimate_pb.Estimate(value=200.0),
                ),
            ],
        ),
        rf_pb.FrequencyOutcomeGrid.ChannelCells(
            channel_name=RF_CHANNELS[1],
            cells=[
                rf_pb.FrequencyOutcomeGrid.Cell(
                    reach_frequency=rf_pb.ReachFrequency(
                        reach=10000,
                        average_frequency=1.0,
                    ),
                    outcome=estimate_pb.Estimate(value=100.0),
                ),
                rf_pb.FrequencyOutcomeGrid.Cell(
                    reach_frequency=rf_pb.ReachFrequency(
                        reach=20000,
                        average_frequency=2.0,
                    ),
                    outcome=estimate_pb.Estimate(value=200.0),
                ),
            ],
        ),
    ],
)

RF_OPTIMIZATION_SPEC_ALL_DATES = rf_pb.ReachFrequencyOptimizationSpec(
    date_interval=ALL_DATE_INTERVAL,
    objective=target_metric_pb.TargetMetric.KPI,
    total_budget_constraint=constraints_pb.BudgetConstraint(
        min_budget=100000.0,
        max_budget=200000.0,
    ),
    rf_channel_constraints=[
        rf_pb.RfChannelConstraint(
            channel_name=RF_CHANNELS[0],
            frequency_constraint=constraints_pb.FrequencyConstraint(
                max_frequency=5.0,
            ),
        ),
        rf_pb.RfChannelConstraint(
            channel_name=RF_CHANNELS[1],
            frequency_constraint=constraints_pb.FrequencyConstraint(
                min_frequency=1.3,
                max_frequency=6.6,
            ),
        ),
    ],
)

RF_OPTIMIZATION_RESULT_FOO = rf_pb.ReachFrequencyOptimizationResult(
    name="reach frequency optimization result foo",
    group_id="group-foo",
    spec=RF_OPTIMIZATION_SPEC_ALL_DATES,
    optimized_channel_frequencies=[
        rf_pb.OptimizedChannelFrequency(
            channel_name=RF_CHANNELS[0],
            optimal_average_frequency=3.3,
        ),
        rf_pb.OptimizedChannelFrequency(
            channel_name=RF_CHANNELS[1],
            optimal_average_frequency=5.6,
        ),
    ],
    optimized_marketing_analysis=ALL_TAG_MARKETING_ANALYSIS_BOTH_OUTCOMES,
    frequency_outcome_grid=FREQUENCY_OUTCOME_GRID_FOO,
)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for time-related operations."""

from collections.abc import Mapping, MutableMapping, Sequence
import datetime

from meridian import constants
from mmm.v1.common import date_interval_pb2
import pandas as pd

from google.type import date_pb2


__all__ = [
    "convert_times_to_date_intervals",
    "create_date_interval_pb",
    "dates_from_date_interval_proto",
]


def convert_times_to_date_intervals(
    times: Sequence[str] | Sequence[datetime.date] | pd.DatetimeIndex,
) -> Mapping[str, date_interval_pb2.DateInterval]:
  """Creates a date interval for each time in `times` as dict values.

  Args:
    times: Sequence of date strings in YYYY-MM-DD format.

  Returns:
    Mapping that maps each time in `times` (string form) to the corresponding
    date interval.

  Raises:
    ValueError: If `times` has fewer than 2 elements or if the interval length
    between each time is not consistent.
  """
  if len(times) < 2:
    raise ValueError("There must be at least 2 time points.")

  if isinstance(times, pd.DatetimeIndex):
    datetimes = [
        datetime.datetime(year=time.year, month=time.month, day=time.day)
        for time in times
    ]
  else:
    datetimes = [
        datetime.datetime.strptime(time, constants.DATE_FORMAT)
        if isinstance(time, str)
        else time
        for time in times
    ]

  interval_length = _compute_interval_length(
      start_date=datetimes[0],
      end_date=datetimes[1],
  )
  time_to_date_interval: MutableMapping[str, date_interval_pb2.DateInterval] = (
      {}
  )

  for i, start_date in enumerate(datetimes):
    if i == len(datetimes) - 1:
      end_date = start_date + datetime.timedelta(days=interval_length)
    else:
      end_date = datetimes[i + 1]
      current_interval_length = _compute_interval_length(
          start_date=start_date,
          end_date=end_date,
      )

      if current_interval_length != interval_length:
        raise ValueError(
            "Interval length between selected times must be consistent."
        )

    date_interval = create_date_interval_pb(start_date, end_date)
    time_to_date_interval[start_date.strftime(constants.DATE_FORMAT)] = (
        date_interval
    )

  return time_to_date_interval


def create_date_interval_pb(
    start_date: datetime.date, end_date: datetime.date, tag: str = ""
) -> date_interval_pb2.DateInterval:
  """Creates a `DateInterval` proto for the given start and end dates.

  Args:
    start_date: A datetime object representing the start date.
    end_date: A datetime object representing the end date.
    tag: An optional tag to identify the date interval.

  Returns:
    Returns a date interval proto wrapping the start/end dates.
  """
  start_date_proto = date_pb2.Date(
      year=start_date.year,
      month=start_date.month,
      day=start_date.day,
  )
  end_date_proto = date_pb2.Date(
      year=end_date.year,
      month=end_date.month,
      day=end_date.day,
  )
  return date_interval_pb2.DateInterval(
      start_date=start_date_proto,
      end_date=end_date_proto,
      tag=tag,
  )


def dates_from_date_interval_proto(
    date_interval: date_interval_pb2.DateInterval,
) -> tuple[datetime.date, datetime.date]:
  """Returns a tuple of `[start, end)` date range from a `DateInterval` proto."""
  start_date = datetime.date(
      date_interval.start_date.year,
      date_interval.start_date.month,
      date_interval.start_date.day,
  )
  end_date = datetime.date(
      date_interval.end_date.year,
      date_interval.end_date.month,
      date_interval.end_date.day,
  )
  return start_date, end_date


def _compute_interval_length(
    start_date: datetime.datetime, end_date: datetime.datetime
) -> int:
  """Computes the number of days between `start_date` and `end_date`.

  Args:
    start_date: A datetime object representing the start date.
    end_date: A datetime object representing the end date.

  Returns:
    The number of days between the given dates.
  """
  return end_date.toordinal() - start_date.toordinal()



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines model specification parameters for Meridian."""

from collections.abc import Mapping
import dataclasses
from typing import Sequence
import warnings

from meridian import constants
from meridian.model import prior_distribution
import numpy as np

__all__ = [
    "ModelSpec",
]


def _validate_roi_calibration_period(
    array: np.ndarray | None,
    array_name: str,
    channel_dim_name: str,
    prior_type: str,
    prior_type_name: str,
):
  """Validates the ROI calibration period array."""
  if array is None:
    return
  if prior_type != constants.TREATMENT_PRIOR_TYPE_ROI:
    raise ValueError(
        f"The `{array_name}` should be `None` unless `{prior_type_name}` is"
        f" '{constants.TREATMENT_PRIOR_TYPE_ROI}'."
    )
  if len(array.shape) != 2:
    raise ValueError(
        f"The shape of the `{array_name}` array {array.shape} should be"
        f" 2-dimensional (`n_media_times` x `{channel_dim_name}`)."
    )


@dataclasses.dataclass(frozen=True)
class ModelSpec:
  """Model specification parameters for Meridian.

  This class contains all model parameters that do not change between the runs
  of Meridian.

  Attributes:
    prior: A `PriorDistribution` object specifying the prior distribution of
      each set of model parameters. The distribution for a vector of parameters
      (for example, `alpha_m`) can be passed as either a scalar distribution or
      a vector distribution. If a scalar distribution is passed, it is broadcast
      to the actual shape of the parameter vector. See `paid_media_prior_type`
      for related details.
    media_effects_dist: A string to specify the distribution of media random
      effects across geos. This attribute is not used with a national-level
      model. Allowed values: `'normal'` or `'log_normal'`. Default:
      `'log_normal'`.
    hill_before_adstock: A boolean indicating whether to apply the Hill function
      before the Adstock function, instead of the default order of Adstock
      before Hill. This argument does not apply to RF channels. Default:
      `False`.
    max_lag: An integer indicating the maximum number of lag periods (≥ `0`) to
      include in the Adstock calculation. Default: `8`.
    unique_sigma_for_each_geo: A boolean indicating whether to use a unique
      residual variance for each geo. If `False`, then a single residual
      variance is used for all geos. Default: `False`.
    media_prior_type: A string to specify the prior type for the media
      coefficients. Allowed values: `'roi'`, `'mroi'`, `'contribution'`,
      `'coefficient'`. The `PriorDistribution` contains `roi_m`, `mroi_m`,
      `contribution_m`, and `beta_m`, but only one of these is used depending on
      the `media_prior_type`. When `media_prior_type` is `'roi'`, the
      `PriorDistribution.roi_m` parameter is used to specify a prior on the ROI.
      When `media_prior_type` is `'mroi'`, the `PriorDistribution.mroi_m`
      parameter is used to specify a prior on the mROI. When `media_prior_type`
      is `'contribution'`, the `PriorDistribution.contribution_m` parameter is
      used to specify a prior on the contribution. When `media_prior_type` is
      `'coefficient'`, the `PriorDistribution.beta_m` parameter is used to
      specify a prior on the coefficient mean parameters. Default: `'roi'`.
    rf_prior_type: A string to specify the prior type for the RF coefficients.
      Allowed values: `'roi'`, `'mroi'`, `'contribution'`, `'coefficient'`. The
      `PriorDistribution` contains distributions `roi_rf`, `mroi_rf`,
      `contribution_rf`, and`beta_rf`, but only one of these is used depending
      on the `rf_prior_type`. When `rf_prior_type` is `'roi'`, the
      `PriorDistribution.roi_rf` parameter is used to specify a prior on the
      ROI. When `rf_media_prior_type` is `'mroi'`, the
      `PriorDistribution.mroi_rf` parameter is used to specify a prior on the
      mROI. When `rf_prior_type` is `'contribution'`, the
      `PriorDistribution.contribution_rf` parameter is used to specify a prior
      on the contribution. When `rf_prior_type` is `'coefficient'`, the
      `PriorDistribution.beta_rf` parameter is used to specify a prior on the
      coefficient mean parameters. Default: `'roi'`.
    paid_media_prior_type: Deprecated. Use `media_prior_type` and
      `rf_prior_type` instead. A string to specify the prior type for media and
      RF treatments at the same time. Ignored when `media_prior_type` or
      `rf_prior_type` are set. Default: `'roi'`.
    roi_calibration_period: An optional boolean array of shape `(n_media_times,
      n_media_channels)` indicating the subset of `time` that the ROI value of
      the `roi_m` prior applies to. The ROI numerator is the incremental outcome
      generated by media executed during the calibration period. More precisely,
      it is the difference in expected outcome between the counterfactual where
      media is set to historical values versus the counterfactual where media is
      set to zero during the calibration period and set to historical values for
      all other time periods. The denominator is the channel spend during
      calibration period (excluding any calibration time periods prior to the
      first KPI time period). Spend data by time period is required. If `None`,
      all times are used. Only used if `media_prior_type` is `'roi'`.
      Default: `None`.
    rf_roi_calibration_period: An optional boolean array of shape
      `(n_media_times, n_rf_channels)` indicating the subset of `time` that the
      ROI value of the `roi_rf` prior applies to. The ROI numerator is the
      incremental outcome generated by media executed during the calibration
      period. More precisely, it is the difference in expected outcome between
      the counterfactual where reach and frequency is set to historical values
      versus the counterfactual where reach is set to zero during the
      calibration period and set to historical values for all other time
      periods. The denominator is the channel spend during calibration period
      (excluding any calibration time periods prior to the first KPI time
      period). Spend data by time period is required. If `None`, all times are
      used. Only used if `rf_prior_type` is `'roi'`. Default: `None`.
    organic_media_prior_type: A string to specify the prior type for the organic
      media coefficients. Allowed values: `'contribution'`, `'coefficient'`.
      `PriorDistribution` contains `contribution_om` and `beta_om`, but only one
      of these is used depending on the `organic_media_prior_type`. When
      `organic_media_prior_type` is `'contribution'`, the
      `PriorDistribution.contribution_om` parameter is used to specify a prior
      on the contribution. When `organic_media_prior_type` is `'coefficient'`,
      the `PriorDistribution.beta_om` parameter is used to specify a prior on
      the coefficient mean parameters. Default: `'contribution'`.
    organic_rf_prior_type: A string to specify the prior type for the organic
      reach and frequency coefficients. Allowed values: `'contribution'`,
      `'coefficient'`. The `PriorDistribution` contains distributions
      `contribution_orf`, and `beta_orf`, but only one of these is used
      depending on the `organic_rf_prior_type`. When `organic_rf_prior_type` is
      `'contribution'`, the `PriorDistribution.contribution_orf` parameter is
      used to specify a prior on the contribution. When `organic_rf_prior_type`
      is `'coefficient'`, the `PriorDistribution.beta_orf` parameter is used to
      specify a prior on the coefficient mean parameters. Default:
      `'contribution'`.
    non_media_treatments_prior_type: A string to specify the prior type for the
      non-media treatment coefficients. Allowed values: `'contribution'`,
      `'coefficient'`. `PriorDistribution` contains `contribution_n` and
      `gamma_n`, but only one of these is used depending on the
      `non_media_prior_type`. When `non_media_prior_type` is `'contribution'`,
      the `PriorDistribution.contribution_n` parameter is used to specify a
      prior on the contribution. When `non_media_prior_type` is `'coefficient'`,
      the `PriorDistribution.gamma_n` parameter is used to specify a prior on
      the coefficient mean parameters. Default: `'contribution'`.
    non_media_baseline_values: Optional list with the shape
      `(n_non_media_channels,)`. Each element is either a float (which means
      that the fixed value will be used as baseline for the given channel) or
      one of the strings `"min"` or `"max"` (which mean that the global minimum
      or maximum value will be used as baseline for the scaled values of the
      given non_media treatments channel). If `None`, the minimum value is used
      as baseline for each non-media treatments channel. This attribute is used
      as the default value for the corresponding argument to `Analyzer` methods.
    knots: An optional integer or list of integers indicating the knots used to
      estimate time effects. When `knots` is a list of integers, the knot
      locations are provided by that list. Zero corresponds to a knot at the
      first time period, one corresponds to a knot at the second time period,
      ..., and `(n_times - 1)` corresponds to a knot at the last time period).
      Typically, we recommend including knots at `0` and `(n_times - 1)`, but
      this is not required. When `knots` is an integer, then there are knots
      with locations equally spaced across the time periods, (including knots at
      zero and `(n_times - 1)`. When `knots` is` 1`, there is a single common
      regression coefficient used for all time periods. If `knots` is set to
      `None`, then the numbers of knots used is equal to the number of time
      periods in the case of a geo model. This is equivalent to each time period
      having its own regression coefficient. If `knots` is set to `None` in the
      case of a national model, then the number of knots used is `1`. Default:
      `None`.
    baseline_geo: An optional integer or a string for the baseline geo. The
      baseline geo is treated as the reference geo in the dummy encoding of
      geos. Non-baseline geos have a corresponding `tau_g` indicator variable,
      meaning that they have a higher prior variance than the baseline geo. When
      set to `None`, the geo with the biggest population is used as the
      baseline. Default: `None`.
    holdout_id: Optional boolean tensor of dimensions `(n_geos, n_times)` for a
      geo-level model or `(n_times,)` for a national model, indicating which
      observations are part of the holdout sample, which are excluded from the
      training sample. Only KPI data is excluded from the training sample. Media
      data is still included as it can affect Adstock for subsequent weeks. If
      "ROI priors" are used, then the `roi_m` parameters correspond to the ROI
      of all geos and times, even those in the holdout sample.
    control_population_scaling_id: An optional boolean tensor of dimension
      `(n_controls,)` indicating the control variables for which the control
      value will be scaled by population. If `None`, no control variables are
      scaled by population. Default: `None`.
    non_media_population_scaling_id: An optional boolean tensor of dimension
      `(n_non_media_channels,)` indicating the non-media variables for which the
      non-media value will be scaled by population. If `None`, then no non-media
      variables are scaled by population. Default: `None`.
    adstock_decay_spec: A string or mapping specifying the adstock decay
      function for each media, RF, organic media and organic RF channel. If a
      string, must be either `'geometric'` or `'binomial'`, specifying that
      decay function for all channels. If a mapping, keys should be channel
      names and values should be `'geometric'` or `'binomial'`, with each
      key-value pair denoting the adstock decay function to use for that
      channel. Channels that are not specified in the mapping default to using
      'geometric'. Default: `'geometric'`.
    enable_aks: A boolean indicating whether to use the Automatic Knot Selection
      algorithm to select an optimal number of knots for running the model
      instead of the default 1 for national models and n_times for geo models.
      If this is set to `True` and the `knots` arg is provided, then an error
      will be raised. Default: `False`.
  """

  prior: prior_distribution.PriorDistribution = dataclasses.field(
      default_factory=prior_distribution.PriorDistribution,
  )
  media_effects_dist: str = constants.MEDIA_EFFECTS_LOG_NORMAL
  hill_before_adstock: bool = False
  max_lag: int = 8
  unique_sigma_for_each_geo: bool = False
  media_prior_type: str | None = None
  rf_prior_type: str | None = None
  paid_media_prior_type: str | None = None
  roi_calibration_period: np.ndarray | None = None
  rf_roi_calibration_period: np.ndarray | None = None
  organic_media_prior_type: str = constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION
  organic_rf_prior_type: str = constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION
  non_media_treatments_prior_type: str = (
      constants.TREATMENT_PRIOR_TYPE_CONTRIBUTION
  )
  non_media_baseline_values: Sequence[float | str] | None = None
  knots: int | list[int] | None = None
  baseline_geo: int | str | None = None
  holdout_id: np.ndarray | None = None
  control_population_scaling_id: np.ndarray | None = None
  non_media_population_scaling_id: np.ndarray | None = None
  adstock_decay_spec: str | Mapping[str, str] = constants.GEOMETRIC_DECAY
  enable_aks: bool = False

  def __post_init__(self):
    # Validate media_effects_dist.
    if self.media_effects_dist not in constants.MEDIA_EFFECTS_DISTRIBUTIONS:
      raise ValueError(
          f"The `media_effects_dist` parameter '{self.media_effects_dist}' must"
          f" be one of {sorted(constants.MEDIA_EFFECTS_DISTRIBUTIONS)}."
      )
    # Support paid_media_prior_type for backwards compatibility.
    if self.paid_media_prior_type is not None:
      if self.media_prior_type is not None or self.rf_prior_type is not None:
        raise ValueError(
            "The deprecated `paid_media_prior_type` parameter cannot be used"
            " with `media_prior_type` or `rf_prior_type`. Use"
            " `media_prior_type` and `rf_prior_type` instead."
        )
      else:
        warnings.warn(
            "Using `paid_media_prior_type` parameter will set prior types for"
            " media and RF at the same time. This is deprecated and will be"
            " removed in a future version of Meridian. Use `media_prior_type`"
            " and `rf_prior_type` instead."
        )
    # Validate prior_type.
    if (
        self.effective_media_prior_type
        not in constants.PAID_TREATMENT_PRIOR_TYPES
    ):
      raise ValueError(
          "The `media_prior_type` parameter"
          f" '{self.effective_media_prior_type}' must be one of"
          f" {sorted(constants.PAID_TREATMENT_PRIOR_TYPES)}."
      )
    if self.effective_rf_prior_type not in constants.PAID_TREATMENT_PRIOR_TYPES:
      raise ValueError(
          "The `rf_prior_type` parameter"
          f" '{self.effective_rf_prior_type}' must be one of"
          f" {sorted(constants.PAID_TREATMENT_PRIOR_TYPES)}."
      )
    if self.organic_media_prior_type not in (
        constants.NON_PAID_TREATMENT_PRIOR_TYPES
    ):
      raise ValueError(
          "The `organic_media_prior_type` parameter"
          f" '{self.organic_media_prior_type}' must be one of"
          f" {sorted(constants.NON_PAID_TREATMENT_PRIOR_TYPES)}."
      )
    if self.organic_rf_prior_type not in (
        constants.NON_PAID_TREATMENT_PRIOR_TYPES
    ):
      raise ValueError(
          "The `organic_rf_prior_type` parameter"
          f" '{self.organic_rf_prior_type}' must be one of"
          f" {sorted(constants.NON_PAID_TREATMENT_PRIOR_TYPES)}."
      )
    if self.non_media_treatments_prior_type not in (
        constants.NON_PAID_TREATMENT_PRIOR_TYPES
    ):
      raise ValueError(
          "The `non_media_treatments_prior_type` parameter"
          f" '{self.non_media_treatments_prior_type}' must be one of"
          f" {sorted(constants.NON_PAID_TREATMENT_PRIOR_TYPES)}."
      )

    _validate_roi_calibration_period(
        array=self.roi_calibration_period,
        array_name="roi_calibration_period",
        channel_dim_name="n_media_channels",
        prior_type=self.effective_media_prior_type,
        prior_type_name="media_prior_type",
    )
    _validate_roi_calibration_period(
        array=self.rf_roi_calibration_period,
        array_name="rf_roi_calibration_period",
        channel_dim_name="n_rf_channels",
        prior_type=self.effective_rf_prior_type,
        prior_type_name="rf_prior_type",
    )

    # Validate knots.
    if isinstance(self.knots, list) and not self.knots:
      raise ValueError("The `knots` parameter cannot be an empty list.")
    if isinstance(self.knots, int) and self.knots == 0:
      raise ValueError("The `knots` parameter cannot be zero.")
    if self.knots is not None and self.enable_aks:
      raise ValueError(
          "The `knots` parameter cannot be set when `enable_aks` is True."
      )

  @property
  def effective_media_prior_type(self) -> str:
    """Returns the effective media prior type.

    The recommended way to set prior types is to use `media_prior_type` and
    `rf_prior_type` directly. If both `media_prior_type` and `rf_prior_type`
    are not set, the deprecated `paid_media_prior_type` is used for both media
    and RF channels. If none of them are set, the default is `roi`.
    """
    if self.media_prior_type is not None:
      return self.media_prior_type
    elif self.paid_media_prior_type is not None:
      return self.paid_media_prior_type
    else:
      return constants.TREATMENT_PRIOR_TYPE_ROI

  @property
  def effective_rf_prior_type(self) -> str:
    """Returns the effective rf prior type.

    The recommended way to set prior types is to use `media_prior_type` and
    `rf_prior_type` directly. If both `media_prior_type` and `rf_prior_type`
    are not set, the deprecated `paid_media_prior_type` is used for both media
    and RF channels. If none of them are set, the default is `roi`.
    """
    if self.rf_prior_type is not None:
      return self.rf_prior_type
    elif self.paid_media_prior_type is not None:
      return self.paid_media_prior_type
    else:
      return constants.TREATMENT_PRIOR_TYPE_ROI



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Converters for `Mmm` protos to flat dataframes.

This package provides a set of tools for transforming data from `Mmm`
protos into flat dataframes. This conversion makes the data easier to analyze,
visualize, and use in other data processing pipelines.
"""

from scenarioplanner.converters.dataframe import budget_opt_converters
from scenarioplanner.converters.dataframe import common
from scenarioplanner.converters.dataframe import constants
from scenarioplanner.converters.dataframe import converter
from scenarioplanner.converters.dataframe import dataframe_model_converter
from scenarioplanner.converters.dataframe import marketing_analyses_converters
from scenarioplanner.converters.dataframe import rf_opt_converters



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for transforming a Meridian model into a structured MMM schema.

This module provides the `ModelKernelProcessor`, which is responsible for
transforming the internal state of a trained Meridian model object into a
structured and portable format defined by the `MmmKernel` protobuf message.

The "kernel" includes essential information about the model, such as:

-   Model specifications and hyperparameters.
-   Inferred parameters distributions (as a serialized ArViz inference data).
-   MMM-agnostic marketing data (i.e. input data to the model).

This serialized representation allows the model to be saved, loaded, and
analyzed across different environments or by other tools that understand the
`MmmKernel` schema.

The serialization logic is primarily handled by the `MeridianSerde` class from
the `schema.serde` package.

Key Classes:

-   `ModelKernelProcessor`: The processor class that takes a Meridian model
    instance and populates an `MmmKernel` message.

Example Usage:

```python
import meridian
from meridian.model import model
from mmm.v1 import mmm_pb2
from schema.processors import model_kernel_processor
import semver

# Assuming 'mmm' is a `meridian.model.Meridian` object.
# Example:
# mmm = meridian.model.Meridian(...)
# mmm.sample_prior(...)
# mmm.sample_posterior(...)

processor = model_kernel_processor.ModelKernelProcessor(
    meridian_model=mmm,
    model_id="my_model_v1",
)

# Create an output Mmm proto message
output_proto = mmm_pb2.Mmm()

# Populate the mmm_kernel field
processor(output_proto)

# Now output_proto.mmm_kernel contains the serialized model.
# This can be saved to a file, sent over a network, etc.
print(f"Model Kernel ID: {output_proto.mmm_kernel.model_id}")
print(f"Meridian Version: {output_proto.mmm_kernel.meridian_version}")
# Access other fields within output_proto.mmm_kernel as needed.
```
"""

import abc

import meridian
from meridian.model import model
from mmm.v1 import mmm_pb2 as pb
from schema.serde import meridian_serde
import semver


class ModelKernelProcessor(abc.ABC):
  """Transcribes a model's stats into an `"MmmKernel` message."""

  def __init__(
      self,
      meridian_model: model.Meridian,
      model_id: str = '',
      meridian_version: semver.VersionInfo = semver.VersionInfo.parse(
          meridian.__version__
      ),
  ):
    """Initializes this `ModelKernelProcessor` with a Meridian model.

    Args:
      meridian_model: A Meridian model.
      model_id: An optional model identifier unique to the given model.
      meridian_version: The version of current Meridian framework.
    """
    self._meridian = meridian_model
    self._model_id = model_id
    self._meridian_version = meridian_version

  def __call__(self, output: pb.Mmm):
    """Sets `mmm_kernel` field in the given `Mmm` proto.

    Args:
      output: The output proto to modify.
    """
    output.mmm_kernel.CopyFrom(
        meridian_serde.MeridianSerde().serialize(
            self._meridian,
            self._model_id,
            self._meridian_version,
            include_convergence_info=True,
        )
    )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Backend Abstraction Layer for Meridian."""

import abc
import functools
import os
from typing import Any, Optional, Sequence, Tuple, TYPE_CHECKING, Union
import warnings

from meridian.backend import config
import numpy as np
from typing_extensions import Literal


# The conditional imports in this module are a deliberate design choice for the
# backend abstraction layer. The TFP-on-JAX substrate provides a nearly
# identical API to the standard TFP library, making an alias-based approach more
# pragmatic than a full Abstract Base Class implementation, which would require
# extensive boilerplate.
# pylint: disable=g-import-not-at-top,g-bad-import-order

_DEFAULT_FLOAT = "float32"
_DEFAULT_INT = "int64"

_TENSORFLOW_TILE_KEYWORD = "multiples"
_JAX_TILE_KEYWORD = "reps"

_ARG_AUTOGRAPH = "autograph"
_ARG_JIT_COMPILE = "jit_compile"
_ARG_STATIC_ARGNUMS = "static_argnums"
_ARG_STATIC_ARGNAMES = "static_argnames"

_DEFAULT_SEED_DTYPE = "int32"
_MAX_INT32 = np.iinfo(np.int32).max

if TYPE_CHECKING:
  import dataclasses
  import jax as _jax
  import tensorflow as _tf

  TensorShapeInstance = Union[_tf.TensorShape, Tuple[int, ...]]

SeedType = Any


def standardize_dtype(dtype: Any) -> str:
  """Converts a backend-specific dtype to a standard string representation.

  Args:
    dtype: A backend-specific dtype object (e.g., tf.DType, np.dtype).

  Returns:
    A canonical string representation of the dtype (e.g., 'float32').
  """

  # Handle None explicitly, as np.dtype(None) defaults to float64.

  if dtype is None:
    return str(None)

  if hasattr(dtype, "as_numpy_dtype"):
    dtype = dtype.as_numpy_dtype

  try:
    return np.dtype(dtype).name
  except TypeError:
    return str(dtype)


def result_type(*types: Any) -> str:
  """Infers the result dtype from a list of input types, backend-agnostically.

  This acts as the single source of truth for type promotion rules. The
  promotion logic is designed to be consistent across all backends.

  Rule: If any input is a float, the result is float32. Otherwise, the result
  is int64 to match NumPy/JAX's default behavior for precision.

  Args:
    *types: A variable number of type objects (e.g., `<class 'int'>`,
      np.dtype('float32')).

  Returns:
    A string representing the promoted dtype.
  """
  standardized_types = []
  for t in types:
    if t is None:
      continue
    try:
      # Standardize the input type before checking promotion rules.
      standardized_types.append(standardize_dtype(t))
    except Exception:  # pylint: disable=broad-except
      # Fallback if standardization fails for an unexpected type.
      standardized_types.append(str(t))

  if any("float" in t for t in standardized_types):
    return _DEFAULT_FLOAT
  return _DEFAULT_INT


def _resolve_dtype(dtype: Optional[Any], *args: Any) -> str:
  """Resolves the final dtype for an operation.

  If a dtype is explicitly provided, it's returned. Otherwise, it infers the
  dtype from the input arguments using the backend-agnostic `result_type`
  promotion rules.

  Args:
    dtype: The user-provided dtype, which may be None.
    *args: The input arguments to the operation, used for dtype inference.

  Returns:
    A string representing the resolved dtype.
  """
  if dtype is not None:
    return standardize_dtype(dtype)

  input_types = [
      getattr(arg, "dtype", type(arg)) for arg in args if arg is not None
  ]
  return result_type(*input_types)


# --- Private Backend-Specific Implementations ---
def _jax_stabilize_rf_roi_grid(
    spend_grid: np.ndarray,
    outcome_grid: np.ndarray,
    n_rf_channels: int,
) -> np.ndarray:
  """Stabilizes the RF ROI grid for JAX using a stable index lookup."""
  new_outcome_grid = outcome_grid.copy()
  rf_slice = slice(-n_rf_channels, None)
  rf_spend_grid = spend_grid[:, rf_slice]
  rf_outcome_grid = new_outcome_grid[:, rf_slice]

  last_valid_indices = np.sum(~np.isnan(rf_spend_grid), axis=0) - 1
  channel_indices = np.arange(n_rf_channels)

  ref_spend = rf_spend_grid[last_valid_indices, channel_indices]
  ref_outcome = rf_outcome_grid[last_valid_indices, channel_indices]

  rf_roi = np.divide(
      ref_outcome,
      ref_spend,
      out=np.zeros_like(ref_outcome, dtype=np.float64),
      where=(ref_spend != 0),
  )

  new_outcome_grid[:, rf_slice] = rf_roi * rf_spend_grid
  return new_outcome_grid


def _tf_stabilize_rf_roi_grid(
    spend_grid: np.ndarray,
    outcome_grid: np.ndarray,
    n_rf_channels: int,
) -> np.ndarray:
  """Stabilizes the RF ROI grid for TF using nanmax logic."""
  new_outcome_grid = outcome_grid.copy()
  rf_slice = slice(-n_rf_channels, None)
  rf_outcome_max = np.nanmax(new_outcome_grid[:, rf_slice], axis=0)
  rf_spend_max = np.nanmax(spend_grid[:, rf_slice], axis=0)

  rf_roi = np.divide(
      rf_outcome_max,
      rf_spend_max,
      out=np.zeros_like(rf_outcome_max, dtype=np.float64),
      where=(rf_spend_max != 0),
  )

  new_outcome_grid[:, rf_slice] = rf_roi * spend_grid[:, rf_slice]
  return new_outcome_grid


def _jax_arange(
    start: Any,
    stop: Optional[Any] = None,
    step: Any = 1,
    dtype: Optional[Any] = None,
) -> "_jax.Array":
  """JAX implementation for arange."""

  # Import locally to make the function self-contained.

  import jax.numpy as jnp

  resolved_dtype = _resolve_dtype(dtype, start, stop, step)
  return jnp.arange(start, stop, step=step, dtype=resolved_dtype)


def _tf_arange(
    start: Any,
    stop: Optional[Any] = None,
    step: Any = 1,
    dtype: Optional[Any] = None,
) -> "_tf.Tensor":
  """TensorFlow implementation for arange."""
  import tensorflow as tf

  resolved_dtype = _resolve_dtype(dtype, start, stop, step)
  try:
    return tf.range(start, limit=stop, delta=step, dtype=resolved_dtype)
  except tf.errors.NotFoundError:
    result = tf.range(start, limit=stop, delta=step, dtype=tf.float32)
    return tf.cast(result, resolved_dtype)


def _jax_cast(x: Any, dtype: Any) -> "_jax.Array":
  """JAX implementation for cast."""
  return jax_ops.asarray(x, dtype=dtype)


def _jax_divide_no_nan(x, y):
  """JAX implementation for divide_no_nan."""
  import jax.numpy as jnp

  return jnp.where(y != 0, jnp.divide(x, y), 0.0)


def _jax_function_wrapper(func=None, **kwargs):
  """A wrapper for jax.jit that handles TF-like args and static args.

  This wrapper provides compatibility with TensorFlow's `tf.function` arguments
  and improves ergonomics when decorating class methods in JAX.

  By default, if neither `static_argnums` nor `static_argnames` are provided, it
  defaults `static_argnums` to `(0,)`. This assumes the function is a method
  where the first argument (`self` or `cls`) should be treated as static.

  To disable this behavior for plain functions, explicitly provide an empty
  tuple: `@backend.function(static_argnums=())`.

  Args:
    func: The function to wrap.
    **kwargs: Keyword arguments passed to jax.jit. TF-specific arguments (like
      `jit_compile`) are ignored.

  Returns:
    The wrapped function or a decorator.
  """
  jit_kwargs = kwargs.copy()

  jit_kwargs.pop(_ARG_JIT_COMPILE, None)
  jit_kwargs.pop(_ARG_AUTOGRAPH, None)

  if _ARG_STATIC_ARGNUMS in jit_kwargs:
    if not jit_kwargs[_ARG_STATIC_ARGNUMS]:
      jit_kwargs.pop(_ARG_STATIC_ARGNUMS)
  else:
    jit_kwargs[_ARG_STATIC_ARGNUMS] = (0,)

  decorator = functools.partial(jax.jit, **jit_kwargs)

  if func:
    return decorator(func)
  return decorator


def _tf_function_wrapper(func=None, **kwargs):
  """A wrapper for tf.function that ignores JAX-specific arguments."""
  import tensorflow as tf

  kwargs.pop(_ARG_STATIC_ARGNAMES, None)
  kwargs.pop(_ARG_STATIC_ARGNUMS, None)

  decorator = tf.function(**kwargs)

  if func:
    return decorator(func)
  return decorator


def _jax_nanmedian(a, axis=None):
  """JAX implementation for nanmedian."""
  import jax.numpy as jnp

  return jnp.nanmedian(a, axis=axis)


def _tf_nanmedian(a, axis=None):
  """TensorFlow implementation for nanmedian using numpy_function."""
  import tensorflow as tf

  return tf.numpy_function(
      lambda x: np.nanmedian(x, axis=axis).astype(x.dtype), [a], a.dtype
  )


def _jax_numpy_function(*args, **kwargs):  # pylint: disable=unused-argument
  raise NotImplementedError(
      "backend.numpy_function is not implemented for the JAX backend."
  )


def _jax_make_tensor_proto(values, dtype=None, shape=None):  # pylint: disable=unused-argument
  """JAX implementation for make_tensor_proto."""
  # pylint: disable=g-direct-tensorflow-import
  from tensorflow.core.framework import tensor_pb2
  from tensorflow.core.framework import tensor_shape_pb2
  from tensorflow.core.framework import types_pb2
  # pylint: enable=g-direct-tensorflow-import

  if not isinstance(values, np.ndarray):
    values = np.array(values)

  if dtype:
    numpy_dtype = np.dtype(dtype)
    values = values.astype(numpy_dtype)
  else:
    numpy_dtype = values.dtype

  dtype_map = {
      np.dtype(np.float16): types_pb2.DT_HALF,
      np.dtype(np.float32): types_pb2.DT_FLOAT,
      np.dtype(np.float64): types_pb2.DT_DOUBLE,
      np.dtype(np.int32): types_pb2.DT_INT32,
      np.dtype(np.uint8): types_pb2.DT_UINT8,
      np.dtype(np.uint16): types_pb2.DT_UINT16,
      np.dtype(np.uint32): types_pb2.DT_UINT32,
      np.dtype(np.uint64): types_pb2.DT_UINT64,
      np.dtype(np.int16): types_pb2.DT_INT16,
      np.dtype(np.int8): types_pb2.DT_INT8,
      np.dtype(np.int64): types_pb2.DT_INT64,
      np.dtype(np.complex64): types_pb2.DT_COMPLEX64,
      np.dtype(np.complex128): types_pb2.DT_COMPLEX128,
      np.dtype(np.bool_): types_pb2.DT_BOOL,
      # Note: String types are handled outside the map.
  }
  proto_dtype = dtype_map.get(numpy_dtype)
  if proto_dtype is None and numpy_dtype.kind in ("S", "U"):
    proto_dtype = types_pb2.DT_STRING

  if proto_dtype is None:
    raise TypeError(
        f"Unsupported dtype for TensorProto conversion: {numpy_dtype}"
    )

  proto = tensor_pb2.TensorProto(
      dtype=proto_dtype,
      tensor_shape=tensor_shape_pb2.TensorShapeProto(
          dim=[
              tensor_shape_pb2.TensorShapeProto.Dim(size=d)
              for d in values.shape
          ]
      ),
  )

  proto.tensor_content = values.tobytes()
  return proto


def _jax_make_ndarray(proto):
  """JAX implementation for make_ndarray."""
  # pylint: disable=g-direct-tensorflow-import
  from tensorflow.core.framework import types_pb2
  # pylint: enable=g-direct-tensorflow-import

  dtype_map = {
      types_pb2.DT_HALF: np.float16,
      types_pb2.DT_FLOAT: np.float32,
      types_pb2.DT_DOUBLE: np.float64,
      types_pb2.DT_INT32: np.int32,
      types_pb2.DT_UINT8: np.uint8,
      types_pb2.DT_UINT16: np.uint16,
      types_pb2.DT_UINT32: np.uint32,
      types_pb2.DT_UINT64: np.uint64,
      types_pb2.DT_INT16: np.int16,
      types_pb2.DT_INT8: np.int8,
      types_pb2.DT_INT64: np.int64,
      types_pb2.DT_COMPLEX64: np.complex64,
      types_pb2.DT_COMPLEX128: np.complex128,
      types_pb2.DT_BOOL: np.bool_,
      types_pb2.DT_STRING: np.bytes_,
  }
  if proto.dtype not in dtype_map:
    raise TypeError(f"Unsupported TensorProto dtype: {proto.dtype}")

  shape = [d.size for d in proto.tensor_shape.dim]
  dtype = dtype_map[proto.dtype]

  if proto.tensor_content:
    num_elements = np.prod(shape).item() if shape else 0
    # When deserializing a string from tensor_content, the itemsize is not
    # explicitly stored. We must infer it from the content length and shape.
    if dtype == np.bytes_ and num_elements > 0:
      content_len = len(proto.tensor_content)
      itemsize = content_len // num_elements
      if itemsize * num_elements != content_len:
        raise ValueError(
            "Tensor content size is not a multiple of the number of elements"
            " for string dtype."
        )
      dtype = np.dtype(f"S{itemsize}")

    return (
        np.frombuffer(proto.tensor_content, dtype=dtype).copy().reshape(shape)
    )

  # Fallback for protos that store data in val fields instead of tensor_content.
  if dtype == np.float32:
    val_field = proto.float_val
  elif dtype == np.float64:
    val_field = proto.double_val
  elif dtype == np.int32:
    val_field = proto.int_val
  elif dtype == np.int64:
    val_field = proto.int64_val
  elif dtype == np.bool_:
    val_field = proto.bool_val
  else:
    if proto.string_val:
      return np.array(proto.string_val, dtype=np.bytes_).reshape(shape)
    if not any(shape):
      return np.array([], dtype=dtype).reshape(shape)
    raise TypeError(
        f"Unsupported dtype for TensorProto value field fallback: {dtype}"
    )

  return np.array(val_field, dtype=dtype).reshape(shape)


def _jax_get_indices_where(condition):
  """JAX implementation for get_indices_where."""
  import jax.numpy as jnp

  return jnp.stack(jnp.where(condition), axis=-1)


def _tf_get_indices_where(condition):
  """TensorFlow implementation for get_indices_where."""
  import tensorflow as tf

  return tf.where(condition)


def _jax_split(value, num_or_size_splits, axis=0):
  """JAX implementation for split that accepts size splits."""
  import jax.numpy as jnp

  if not isinstance(num_or_size_splits, int):
    indices = jnp.cumsum(jnp.array(num_or_size_splits))[:-1]
    return jnp.split(value, indices, axis=axis)

  return jnp.split(value, num_or_size_splits, axis=axis)


def _jax_tile(*args, **kwargs):
  """JAX wrapper for tile that supports the `multiples` keyword argument."""
  import jax.numpy as jnp

  if _TENSORFLOW_TILE_KEYWORD in kwargs:
    kwargs[_JAX_TILE_KEYWORD] = kwargs.pop(_TENSORFLOW_TILE_KEYWORD)
  return jnp.tile(*args, **kwargs)


def _jax_unique_with_counts(x):
  """JAX implementation for unique_with_counts."""
  import jax.numpy as jnp

  y, counts = jnp.unique(x, return_counts=True)
  # The TF version returns a tuple of (y, idx, count). The idx is not used in
  # the calling code, so we can return None for it to maintain tuple structure.
  return y, None, counts


def _tf_unique_with_counts(x):
  """TensorFlow implementation for unique_with_counts."""
  import tensorflow as tf

  return tf.unique_with_counts(x)


def _jax_boolean_mask(tensor, mask, axis=None):
  """JAX implementation for boolean_mask that supports an axis argument."""
  import jax.numpy as jnp

  if axis is None:
    axis = 0
  mask = jnp.asarray(mask)
  tensor_swapped = jnp.moveaxis(tensor, axis, 0)
  masked = tensor_swapped[mask]
  return jnp.moveaxis(masked, 0, axis)


def _tf_boolean_mask(tensor, mask, axis=None):
  """TensorFlow implementation for boolean_mask."""
  import tensorflow as tf

  return tf.boolean_mask(tensor, mask, axis=axis)


def _jax_gather(params, indices, axis=0):
  """JAX implementation for gather with axis support."""
  import jax.numpy as jnp

  if isinstance(params, (list, tuple)):
    params = np.array(params)

  # JAX can't JIT-compile operations on string or object arrays. We detect
  # these types and fall back to standard NumPy operations.
  if isinstance(params, np.ndarray) and params.dtype.kind in ("S", "U", "O"):
    return np.take(params, np.asarray(indices), axis=axis)

  return jnp.take(params, indices, axis=axis)


def _tf_gather(params, indices, axis=0):
  """TensorFlow implementation for gather."""
  import tensorflow as tf

  return tf.gather(params, indices, axis=axis)


def _jax_fill(dims, value):
  """JAX implementation for fill."""
  import jax.numpy as jnp

  return jnp.full(dims, value)


def _tf_fill(dims, value):
  """TensorFlow implementation for fill."""
  import tensorflow as tf

  return tf.fill(dims, value)


def _jax_argmax(tensor, axis=None):
  """JAX implementation for argmax, aligned with TensorFlow's default.

  This function finds the indices of the maximum values along a specified axis.
  Crucially, it mimics the default behavior of TensorFlow's `tf.argmax`, where
  if `axis` is `None`, the operation defaults to `axis=0`. This differs from
  NumPy's and JAX's native `argmax` behavior, which would flatten the array
  before finding the index.

  Args:
    tensor: The input JAX array.
    axis: An integer specifying the axis along which to find the index of the
      maximum value. If `None`, it defaults to `0` to match TensorFlow's
      behavior.

  Returns:
    A JAX array containing the indices of the maximum values.
  """
  import jax.numpy as jnp

  if axis is None:
    axis = 0
  return jnp.argmax(tensor, axis=axis)


def _tf_argmax(tensor, axis=None):
  """TensorFlow implementation for argmax."""
  import tensorflow as tf

  return tf.argmax(tensor, axis=axis)


def _jax_broadcast_dynamic_shape(shape_x, shape_y):
  """JAX implementation for broadcast_dynamic_shape."""
  import jax.numpy as jnp

  return jnp.broadcast_shapes(shape_x, shape_y)


def _jax_tensor_shape(dims):
  """JAX implementation for TensorShape."""
  if isinstance(dims, int):
    return (dims,)

  return tuple(dims)


def _jax_transpose(a, perm=None):
  """JAX wrapper for transpose to support the 'perm' keyword argument."""
  import jax.numpy as jnp

  return jnp.transpose(a, axes=perm)


def _jax_get_seed_data(seed: Any) -> Optional[np.ndarray]:
  """Extracts the underlying numerical data from a JAX PRNGKey."""
  if seed is None:
    return None

  return np.array(jax.random.key_data(seed))


def _tf_get_seed_data(seed: Any) -> Optional[np.ndarray]:
  """Converts a TensorFlow-style seed into a NumPy array."""
  if seed is None:
    return None
  return np.array(seed)


def _jax_convert_to_tensor(data, dtype=None):
  """Converts data to a JAX array, handling strings as NumPy arrays.

  This function explicitly unwraps objects with a `.values` attribute (e.g.,
  pandas.DataFrame, xarray.DataArray) to access the underlying NumPy array,
  provided that `.values` is not a method. This takes precedence over the
  `__array__` protocol.

  It also handles precision mismatches: if `data` is float64 and `dtype` is
  not specified, and JAX x64 mode is disabled (default), it issues a warning
  and explicitly casts to float32 to match the backend default and prevent
  silent precision loss or type errors in downstream operations.

  Args:
    data: The data to convert.
    dtype: The desired data type.

  Returns:
    A JAX array, or a NumPy array if the dtype is a string type.
  """
  # Unwrap xarray.DataArray, pandas.Series, and pandas.DataFrame objects.
  # These objects wrap the underlying NumPy array in a .values attribute.
  if hasattr(data, "values") and not callable(data.values):
    data = data.values

  # Convert to numpy array upfront to simplify dtype inspection below.
  # A standard Python float is 64-bit, and this conversion allows the
  # subsequent logic to correctly detect and handle potential float64
  # downcasting for scalar inputs.
  if isinstance(data, (list, tuple, float)):
    data = np.array(data)

  # JAX does not natively support string tensors in the same way TF does.
  # If a string dtype is requested, or if the data is inherently strings,
  # we fall back to a standard NumPy array.
  is_string_target = False
  if dtype is not None:
    try:
      if np.dtype(dtype).kind in ("S", "U"):
        is_string_target = True
    except TypeError:
      # This can happen if dtype is not a valid dtype specifier,
      # let jax.asarray handle it.
      pass

  is_string_data = isinstance(data, np.ndarray) and data.dtype.kind in (
      "S",
      "U",
  )

  if is_string_target:
    return np.array(data, dtype=dtype)

  if dtype is None and is_string_data:
    return data

  # If the user provides float64 data but does not request a specific dtype,
  # and JAX 64-bit mode is disabled (default), JAX would implicitly truncate.
  # We cast to float32 and warn the user to prevent silent mismatches.
  if dtype is None:
    is_float64_input = hasattr(data, "dtype") and data.dtype == np.float64
    if is_float64_input:
      if not jax.config.jax_enable_x64:
        warnings.warn(
            "Input data is float64. Casting to float32 to match backend "
            "default precision.",
            UserWarning,
        )
        dtype = jax_ops.float32

  return jax_ops.asarray(data, dtype=dtype)


def _tf_nanmean(a, axis=None, keepdims=False):
  import tensorflow.experimental.numpy as tnp

  return tf_backend.convert_to_tensor(
      tnp.nanmean(a, axis=axis, keepdims=keepdims)
  )


def _tf_nansum(a, axis=None, keepdims=False):
  import tensorflow.experimental.numpy as tnp

  return tf_backend.convert_to_tensor(
      tnp.nansum(a, axis=axis, keepdims=keepdims)
  )


def _tf_nanvar(a, axis=None, keepdims=False):
  """Calculates variance ignoring NaNs, strictly returning a Tensor."""
  import tensorflow as tf
  import tensorflow.experimental.numpy as tnp
  # We implement two-pass variance to correctly handle NaNs and ensure
  # all operations remain within the TF graph (maintaining differentiability).
  a_tensor = tf.convert_to_tensor(a)
  mean = tnp.nanmean(a_tensor, axis=axis, keepdims=True)
  sq_diff = tf.math.squared_difference(a_tensor, mean)
  var = tnp.nanmean(sq_diff, axis=axis, keepdims=keepdims)
  return tf.convert_to_tensor(var)


def _jax_one_hot(
    indices, depth, on_value=None, off_value=None, axis=None, dtype=None
):
  """JAX implementation for one_hot."""
  import jax.numpy as jnp

  resolved_dtype = _resolve_dtype(dtype, on_value, off_value, 1, 0)
  jax_axis = -1 if axis is None else axis

  one_hot_result = jax.nn.one_hot(
      indices, num_classes=depth, dtype=jnp.dtype(resolved_dtype), axis=jax_axis
  )

  on_val = 1 if on_value is None else on_value
  off_val = 0 if off_value is None else off_value

  if on_val == 1 and off_val == 0:
    return one_hot_result

  on_tensor = jnp.array(on_val, dtype=jnp.dtype(resolved_dtype))
  off_tensor = jnp.array(off_val, dtype=jnp.dtype(resolved_dtype))

  return jnp.where(one_hot_result == 1, on_tensor, off_tensor)


def _jax_roll(a, shift, axis=None):
  """JAX implementation for roll."""
  import jax.numpy as jnp

  return jnp.roll(a, shift, axis=axis)


def _tf_roll(a, shift: Sequence[int], axis=None):
  """TensorFlow implementation for roll that handles axis=None."""
  import tensorflow as tf

  if axis is None:
    original_shape = tf.shape(a)
    flat_tensor = tf.reshape(a, [-1])
    rolled_flat = tf.roll(flat_tensor, shift=shift, axis=0)
    return tf.reshape(rolled_flat, original_shape)
  return tf.roll(a, shift, axis=axis)


def _jax_enable_op_determinism():
  """No-op for JAX. Determinism is handled via stateless PRNGKeys."""
  warnings.warn(
      "op determinism is a TensorFlow-specific concept and has no effect when"
      " using the JAX backend."
  )


# --- Backend Initialization ---
_BACKEND = config.get_backend()

# We expose standardized functions directly at the module level (backend.foo)
# to provide a consistent, NumPy-like API across backends. The '_ops' object
# is a private member for accessing the full, raw backend library if necessary,
# but usage should prefer the top-level standardized functions.

if _BACKEND == config.Backend.JAX:
  import jax
  import jax.numpy as jax_ops
  import tensorflow_probability.substrates.jax as tfp_jax
  from jax import tree_util

  class ExtensionType:
    """A JAX-compatible stand-in for tf.experimental.ExtensionType.

    This class registers itself as a JAX Pytree node, allowing it to be passed
    through JIT-compiled functions.
    """

    def __init_subclass__(cls, **kwargs):
      super().__init_subclass__(**kwargs)
      tree_util.register_pytree_node(
          cls,
          cls._tree_flatten,
          cls._tree_unflatten,
      )

    def _tree_flatten(self):
      """Flattens the object for JAX tracing.

      Fields containing JAX arrays (or convertibles) are treated as children.
      Fields containing strings or NumPy string arrays must be treated as
      auxiliary data because JAX cannot trace non-numeric/non-boolean data.

      Returns:
        A tuple of (children, aux_data), where children are the
        tracer-compatible parts of the object, and aux_data contains auxiliary
        information needed for unflattening.
      """
      d = vars(self)
      all_keys = sorted(d.keys())
      children = []
      aux = {}
      children_keys = []

      for k in all_keys:
        v = d[k]
        # Identify string data to prevent JAX tracing errors.
        # 'S' is zero-terminated bytes (fixed-width), 'U' is unicode string.
        is_numpy_string = isinstance(v, np.ndarray) and v.dtype.kind in (
            "S",
            "U",
        )
        is_plain_string = isinstance(v, str)

        if is_numpy_string or is_plain_string:
          aux[k] = v
        else:
          children.append(v)
          children_keys.append(k)
      return children, (aux, children_keys)

    @classmethod
    def _tree_unflatten(cls, aux_and_keys, children):
      aux, children_keys = aux_and_keys
      obj = cls.__new__(cls)
      vars(obj).update(aux)
      for k, v in zip(children_keys, children):
        setattr(obj, k, v)
      return obj

  class _JaxErrors:
    # pylint: disable=invalid-name
    ResourceExhaustedError = MemoryError
    InvalidArgumentError = ValueError
    # pylint: enable=invalid-name

  class _JaxRandom:
    """Provides JAX-based random number generation utilities.

    This class mirrors the structure needed by `RNGHandler` for JAX.
    """

    @staticmethod
    def prng_key(seed):
      return jax.random.PRNGKey(seed)

    @staticmethod
    def split(key):
      return jax.random.split(key)

    @staticmethod
    def generator_from_seed(seed):
      raise NotImplementedError("JAX backend does not use Generators.")

    @staticmethod
    def stateless_split(seed: Any, num: int = 2):
      raise NotImplementedError(
          "Direct stateless splitting from an integer seed is not the primary"
          " pattern used in the JAX backend. Use `backend.random.split(key)`"
          " instead."
      )

    @staticmethod
    def stateless_randint(key, shape, minval, maxval, dtype=jax_ops.int32):
      """Wrapper for jax.random.randint."""
      return jax.random.randint(
          key, shape, minval=minval, maxval=maxval, dtype=dtype
      )

    @staticmethod
    def stateless_uniform(
        key, shape, dtype=jax_ops.float32, minval=0.0, maxval=1.0
    ):
      """Replacement for tfp_jax.random.stateless_uniform using jax.random.uniform."""
      return jax.random.uniform(
          key, shape=shape, dtype=dtype, minval=minval, maxval=maxval
      )

    @staticmethod
    def sanitize_seed(seed):
      return tfp_jax.random.sanitize_seed(seed)

  random = _JaxRandom()

  @functools.partial(
      jax.jit,
      static_argnames=[
          "joint_dist",
          "n_chains",
          "n_draws",
          "num_adaptation_steps",
          "dual_averaging_kwargs",
          "max_tree_depth",
          "unrolled_leapfrog_steps",
          "parallel_iterations",
      ],
  )
  def _jax_xla_windowed_adaptive_nuts(**kwargs):
    """JAX-specific JIT wrapper for the NUTS sampler."""
    kwargs["seed"] = random.prng_key(kwargs["seed"])
    return experimental.mcmc.windowed_adaptive_nuts(**kwargs)

  xla_windowed_adaptive_nuts = _jax_xla_windowed_adaptive_nuts

  _ops = jax_ops
  errors = _JaxErrors()
  Tensor = jax.Array
  tfd = tfp_jax.distributions
  bijectors = tfp_jax.bijectors
  experimental = tfp_jax.experimental
  mcmc = tfp_jax.mcmc
  _convert_to_tensor = _jax_convert_to_tensor

  # Standardized Public API
  absolute = _ops.abs
  allclose = _ops.allclose
  arange = _jax_arange
  argmax = _jax_argmax
  boolean_mask = _jax_boolean_mask
  broadcast_dynamic_shape = _jax_broadcast_dynamic_shape
  broadcast_to = _ops.broadcast_to
  cast = _jax_cast
  concatenate = _ops.concatenate
  cumsum = _ops.cumsum
  divide = _ops.divide
  divide_no_nan = _jax_divide_no_nan
  einsum = _ops.einsum
  enable_op_determinism = _jax_enable_op_determinism
  equal = _ops.equal
  exp = _ops.exp
  expand_dims = _ops.expand_dims
  fill = _jax_fill
  function = _jax_function_wrapper
  gather = _jax_gather
  get_indices_where = _jax_get_indices_where
  get_seed_data = _jax_get_seed_data
  is_nan = _ops.isnan
  log = _ops.log
  make_ndarray = _jax_make_ndarray
  make_tensor_proto = _jax_make_tensor_proto
  nanmean = jax_ops.nanmean
  nanmedian = _jax_nanmedian
  nansum = jax_ops.nansum
  nanvar = jax_ops.nanvar
  numpy_function = _jax_numpy_function
  one_hot = _jax_one_hot
  ones = _ops.ones
  ones_like = _ops.ones_like
  rank = _ops.ndim
  reduce_any = _ops.any
  reduce_max = _ops.max
  reduce_mean = _ops.mean
  reduce_min = _ops.min
  reduce_std = _ops.std
  reduce_sum = _ops.sum
  repeat = _ops.repeat
  reshape = _ops.reshape
  roll = _jax_roll
  split = _jax_split
  stack = _ops.stack
  tile = _jax_tile
  transpose = _jax_transpose
  unique_with_counts = _jax_unique_with_counts
  where = _ops.where
  zeros = _ops.zeros
  zeros_like = _ops.zeros_like

  float32 = _ops.float32
  bool_ = _ops.bool_
  newaxis = _ops.newaxis
  TensorShape = _jax_tensor_shape
  int32 = _ops.int32
  string = np.bytes_

  stabilize_rf_roi_grid = _jax_stabilize_rf_roi_grid

  def set_random_seed(seed: int) -> None:  # pylint: disable=unused-argument
    warnings.warn(
        "backend.set_random_seed is a no-op in JAX. Randomness is managed "
        "statelessly via PRNGKeys passed to sampling methods."
    )

elif _BACKEND == config.Backend.TENSORFLOW:
  import tensorflow as tf_backend
  import tensorflow_probability as tfp

  _ops = tf_backend
  errors = _ops.errors

  Tensor = tf_backend.Tensor
  ExtensionType = _ops.experimental.ExtensionType

  class _TfRandom:
    """Provides TensorFlow-based random number generation utilities.

    This class mirrors the structure needed by `RNGHandler` for TensorFlow.
    """

    @staticmethod
    def prng_key(seed):
      raise NotImplementedError(
          "TensorFlow backend does not use explicit PRNG keys for"
          " standard sampling."
      )

    @staticmethod
    def split(key):
      raise NotImplementedError(
          "TensorFlow backend does not implement explicit key splitting."
      )

    @staticmethod
    def generator_from_seed(seed):
      return tf_backend.random.Generator.from_seed(seed)

    @staticmethod
    def stateless_split(
        seed: "tf_backend.Tensor", num: int = 2
    ) -> "tf_backend.Tensor":
      return tf_backend.random.experimental.stateless_split(seed, num=num)

    @staticmethod
    def stateless_randint(
        seed: "tf_backend.Tensor",
        shape: "TensorShapeInstance",
        minval: int,
        maxval: int,
        dtype: Any = _DEFAULT_SEED_DTYPE,
    ) -> "tf_backend.Tensor":
      return tf_backend.random.stateless_uniform(
          shape=shape, seed=seed, minval=minval, maxval=maxval, dtype=dtype
      )

    @staticmethod
    def stateless_uniform(
        seed, shape, minval=0, maxval=None, dtype=tf_backend.float32
    ):
      return tf_backend.random.stateless_uniform(
          shape=shape, seed=seed, minval=minval, maxval=maxval, dtype=dtype
      )

    @staticmethod
    def sanitize_seed(seed):
      return tfp.random.sanitize_seed(seed)

  random = _TfRandom()

  @tf_backend.function(autograph=False, jit_compile=True)
  def _tf_xla_windowed_adaptive_nuts(**kwargs):
    """TensorFlow-specific XLA wrapper for the NUTS sampler."""
    return experimental.mcmc.windowed_adaptive_nuts(**kwargs)

  xla_windowed_adaptive_nuts = _tf_xla_windowed_adaptive_nuts

  tfd = tfp.distributions
  bijectors = tfp.bijectors
  experimental = tfp.experimental
  mcmc = tfp.mcmc
  _convert_to_tensor = _ops.convert_to_tensor

  # Standardized Public API
  absolute = _ops.math.abs
  allclose = _ops.experimental.numpy.allclose
  arange = _tf_arange
  argmax = _tf_argmax
  boolean_mask = _tf_boolean_mask
  broadcast_dynamic_shape = _ops.broadcast_dynamic_shape
  broadcast_to = _ops.broadcast_to
  cast = _ops.cast
  concatenate = _ops.concat
  cumsum = _ops.cumsum
  divide = _ops.divide
  divide_no_nan = _ops.math.divide_no_nan
  einsum = _ops.einsum
  enable_op_determinism = _ops.config.experimental.enable_op_determinism
  equal = _ops.equal
  exp = _ops.math.exp
  expand_dims = _ops.expand_dims
  fill = _tf_fill
  function = _tf_function_wrapper
  gather = _tf_gather
  get_indices_where = _tf_get_indices_where
  get_seed_data = _tf_get_seed_data
  is_nan = _ops.math.is_nan
  log = _ops.math.log
  make_ndarray = _ops.make_ndarray
  make_tensor_proto = _ops.make_tensor_proto
  nanmean = _tf_nanmean
  nanmedian = _tf_nanmedian
  nansum = _tf_nansum
  nanvar = _tf_nanvar
  numpy_function = _ops.numpy_function
  one_hot = _ops.one_hot
  ones = _ops.ones
  ones_like = _ops.ones_like
  rank = _ops.rank
  reduce_any = _ops.reduce_any
  reduce_max = _ops.reduce_max
  reduce_mean = _ops.reduce_mean
  reduce_min = _ops.reduce_min
  reduce_std = _ops.math.reduce_std
  reduce_sum = _ops.reduce_sum
  repeat = _ops.repeat
  reshape = _ops.reshape
  roll = _tf_roll
  set_random_seed = tf_backend.keras.utils.set_random_seed
  split = _ops.split
  stack = _ops.stack
  tile = _ops.tile
  transpose = _ops.transpose
  unique_with_counts = _tf_unique_with_counts
  where = _ops.where
  zeros = _ops.zeros
  zeros_like = _ops.zeros_like

  stabilize_rf_roi_grid = _tf_stabilize_rf_roi_grid

  float32 = _ops.float32
  bool_ = _ops.bool
  newaxis = _ops.newaxis
  TensorShape = _ops.TensorShape
  int32 = _ops.int32
  string = _ops.string

else:
  raise ValueError(f"Unsupported backend: {_BACKEND}")
# pylint: enable=g-import-not-at-top,g-bad-import-order


def _extract_int_seed(s: Any) -> Optional[int]:
  """Attempts to extract a scalar Python integer from various input types.

  Args:
    s: The input seed, which can be an int, Tensor, or array-like object.

  Returns:
    A Python integer if a scalar integer can be extracted, otherwise None.
  """
  try:
    if isinstance(s, int):
      return s

    value = np.asarray(s)

    if value.ndim == 0 and np.issubdtype(value.dtype, np.integer):
      return int(value)
  # A broad exception is used here because the input `s` can be of many types
  # (e.g., JAX PRNGKey) which may cause np.asarray or other operations to fail
  # in unpredictable ways. The goal is to safely attempt extraction and fail
  # gracefully.
  except Exception:  # pylint: disable=broad-except
    return None
  return None


class _BaseRNGHandler(abc.ABC):
  """A backend-agnostic abstract base class for random number generation state.

  This handler provides a stateful-style interface for consuming randomness,
  abstracting away the differences between JAX's stateless PRNG keys and
  TensorFlow's paradigms.

  Attributes:
    _seed_input: The original seed object provided during initialization.
    _int_seed: A Python integer extracted from the seed, if possible.
  """

  def __init__(self, seed: SeedType):
    """Initializes the RNG handler.

    Args:
      seed: The initial seed. The accepted type depends on the backend. For JAX,
        this must be an integer. For TensorFlow, this can be an integer, a
        sequence of two integers, or a Tensor. If None, the handler becomes a
        no-op, returning None for all seed requests.
    """
    self._seed_input = seed
    self._int_seed: Optional[int] = _extract_int_seed(seed)

  @abc.abstractmethod
  def get_kernel_seed(self) -> Any:
    """Provides a backend-appropriate sanitized seed/key for an MCMC kernel.

    This method exposes the current state of the handler in the format expected
    by the backend's MCMC machinery. It does not advance the internal state.

    Returns:
      A backend-specific seed object (e.g., a JAX PRNGKey or a TF Tensor).
    """
    raise NotImplementedError

  @abc.abstractmethod
  def get_next_seed(self) -> Any:
    """Provides the appropriate seed object for the next sequential operation.

    This is primarily used for prior sampling and typically advances the
    internal state.

    Returns:
      A backend-specific seed object for a single random operation.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def advance_handler(self) -> "_BaseRNGHandler":
    """Creates a new, independent RNGHandler for a subsequent operation.

    This method is used to generate a new handler derived deterministically
    from the current handler's state.

    Returns:
      A new, independent `RNGHandler` instance.
    """
    raise NotImplementedError


class _JaxRNGHandler(_BaseRNGHandler):
  """JAX implementation of the RNGHandler using explicit key splitting."""

  def __init__(self, seed: SeedType):
    """Initializes the JAX RNG handler.

    Args:
      seed: The initial seed, which must be a Python integer, a scalar integer
        Tensor/array, or None.

    Raises:
      ValueError: If the provided seed is not a scalar integer or None.
    """
    super().__init__(seed)
    self._key: Optional["_jax.Array"] = None

    if seed is None:
      return

    if (
        isinstance(seed, jax.Array)  # pylint: disable=undefined-variable
        and seed.shape == (2,)
        and seed.dtype == jax_ops.uint32  # pylint: disable=undefined-variable
    ):
      self._key = seed
      return

    if self._int_seed is None:
      raise ValueError(
          "JAX backend requires a seed that is an integer or a scalar array,"
          f" but got: {type(seed)} with value {seed!r}"
      )

    self._key = random.prng_key(self._int_seed)

  def get_kernel_seed(self) -> Any:
    if self._key is None:
      return None
    _, subkey = random.split(self._key)
    return int(jax.random.randint(subkey, (), 0, 2**31 - 1))  # pylint: disable=undefined-variable

  def get_next_seed(self) -> Any:
    if self._key is None:
      return None
    self._key, subkey = random.split(self._key)
    return subkey

  def advance_handler(self) -> "_JaxRNGHandler":
    if self._key is None:
      return _JaxRNGHandler(None)

    self._key, subkey_for_new_handler = random.split(self._key)
    new_seed_tensor = random.stateless_randint(
        key=subkey_for_new_handler,
        shape=(),
        minval=0,
        maxval=_MAX_INT32,
        dtype=_DEFAULT_SEED_DTYPE,
    )
    return _JaxRNGHandler(np.asarray(new_seed_tensor).item())


class _TFRNGHandler(_BaseRNGHandler):
  """A stateless-style RNG handler for TensorFlow.

  This handler canonicalizes any seed input into a single stateless seed Tensor.
  """

  def __init__(self, seed: SeedType):
    """Initializes the TensorFlow RNG handler.

    Args:
      seed: The initial seed. Can be an integer, a sequence of two integers, a
        corresponding Tensor, or None. It will be sanitized and stored
        internally as a single stateless seed Tensor.
    """
    super().__init__(seed)
    self._seed_state: Optional["_tf.Tensor"] = None

    if seed is None:
      return

    if isinstance(seed, Sequence) and len(seed) != 2:
      raise ValueError(
          "Invalid seed: Must be either a single integer (stateful seed) or a"
          " pair of two integers (stateless seed). See"
          " [tfp.random.sanitize_seed](https://www.tensorflow.org/probability/api_docs/python/tfp/random/sanitize_seed)"
          " for details."
      )

    if isinstance(seed, int):
      seed_to_sanitize = (seed, seed)
    else:
      seed_to_sanitize = seed

    self._seed_state = random.sanitize_seed(seed_to_sanitize)

  def get_kernel_seed(self) -> Any:
    """Provides the current stateless seed of the handler without advancing it."""
    return self._seed_state

  def get_next_seed(self) -> Any:
    """Returns a new unique seed and advances the handler's internal state."""
    if self._seed_state is None:
      return None
    new_seed, next_state = random.stateless_split(self._seed_state)
    self._seed_state = next_state
    return new_seed

  def advance_handler(self) -> "_TFRNGHandler":
    """Creates a new handler from a new seed and advances the current handler."""
    if self._seed_state is None:
      return _TFRNGHandler(None)
    seed_for_new_handler, next_state = random.stateless_split(self._seed_state)
    self._seed_state = next_state
    return _TFRNGHandler(seed_for_new_handler)


if _BACKEND == config.Backend.JAX:
  RNGHandler = _JaxRNGHandler
elif _BACKEND == config.Backend.TENSORFLOW:
  RNGHandler = _TFRNGHandler
else:
  raise ImportError(f"RNGHandler not implemented for backend: {_BACKEND}")


def to_tensor(data: Any, dtype: Optional[Any] = None) -> Tensor:  # type: ignore
  """Converts input data to the currently active backend tensor type.

  Args:
    data: The data to convert.
    dtype: The desired data type of the resulting tensor. The accepted types
      depend on the active backend (e.g., jax.numpy.dtype or tf.DType).

  Returns:
    A tensor representation of the data for the active backend.
  """

  return _convert_to_tensor(data, dtype=dtype)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common testing utilities for Meridian, designed to be backend-agnostic."""

import dataclasses
from typing import Any, Optional

from absl.testing import parameterized
from google.protobuf import descriptor
from google.protobuf import message
from meridian import backend
from meridian.backend import config
import numpy as np

from tensorflow.python.util.protobuf import compare
# pylint: disable=g-direct-tensorflow-import
from tensorflow.core.framework import tensor_pb2


# pylint: enable=g-direct-tensorflow-import

FieldDescriptor = descriptor.FieldDescriptor

# A type alias for backend-agnostic array-like objects.
# We use `Any` here to avoid circular dependencies with the backend module
# while still allowing the function to accept backend-specific tensor types.
ArrayLike = Any


def assert_allclose(
    a: ArrayLike,
    b: ArrayLike,
    rtol: float = 1e-6,
    atol: float = 1e-6,
    err_msg: str = "",
):
  """Backend-agnostic assertion to check if two array-like objects are close.

  This function converts both inputs to NumPy arrays before comparing them,
  making it compatible with TensorFlow Tensors, JAX Arrays, and standard
  Python lists or NumPy arrays.

  Args:
    a: The first array-like object to compare.
    b: The second array-like object to compare.
    rtol: The relative tolerance parameter.
    atol: The absolute tolerance parameter.
    err_msg: The error message to be printed in case of failure.

  Raises:
    AssertionError: If the two arrays are not equal within the given tolerance.
  """
  np.testing.assert_allclose(
      np.array(a), np.array(b), rtol=rtol, atol=atol, err_msg=err_msg
  )


def assert_allequal(a: ArrayLike, b: ArrayLike, err_msg: str = ""):
  """Backend-agnostic assertion to check if two array-like objects are equal.

  This function converts both inputs to NumPy arrays before comparing them.

  Args:
    a: The first array-like object to compare.
    b: The second array-like object to compare.
    err_msg: The error message to be printed in case of failure.

  Raises:
    AssertionError: If the two arrays are not equal.
  """
  np.testing.assert_array_equal(np.array(a), np.array(b), err_msg=err_msg)


def assert_deep_equals(
    test_case,
    obj1: Any,
    obj2: Any,
    msg: str = "",
    rtol: float = 1e-5,
    atol: float = 1e-5,
):
  """Recursive equality check handling Dataclasses, Lists, and Backend Tensors.

  Args:
    test_case: The unittest.TestCase instance (self) to use for assertions.
    obj1: The first object to compare.
    obj2: The second object to compare.
    msg: Optional error message prefix.
    rtol: Relative tolerance for float comparison.
    atol: Absolute tolerance for float comparison.
  """
  if obj1 is None or obj2 is None:
    test_case.assertEqual(obj1, obj2, msg=msg)
    return

  if (
      hasattr(obj1, "__array__")
      or hasattr(obj1, "numpy")
      or isinstance(obj1, (np.ndarray, backend.Tensor))
  ):
    arr1 = np.array(obj1)
    arr2 = np.array(obj2)

    # Check for non-numeric types where atol/rtol don't apply
    if arr1.dtype.kind in ("U", "S", "O", "b"):
      np.testing.assert_array_equal(arr1, arr2, err_msg=msg)
    else:
      np.testing.assert_allclose(arr1, arr2, err_msg=msg, rtol=rtol, atol=atol)
    return

  if dataclasses.is_dataclass(obj1):
    test_case.assertIs(
        type(obj1),
        type(obj2),
        msg=f"{msg} Type mismatch: {type(obj1)} vs {type(obj2)}",
    )
    for field in dataclasses.fields(obj1):
      val1 = getattr(obj1, field.name)
      val2 = getattr(obj2, field.name)
      assert_deep_equals(
          test_case,
          val1,
          val2,
          msg=f"{msg}.{field.name}",
          rtol=rtol,
          atol=atol,
      )
    return

  if isinstance(obj1, (list, tuple)):
    test_case.assertIsInstance(obj2, (list, tuple), msg=f"{msg} Type mismatch")
    test_case.assertEqual(len(obj1), len(obj2), msg=f"{msg} Length mismatch")
    for i, (item1, item2) in enumerate(zip(obj1, obj2)):
      assert_deep_equals(
          test_case, item1, item2, msg=f"{msg}[{i}]", rtol=rtol, atol=atol
      )
    return

  # Fallback to standard equality for primitives (int, str, float, etc.)
  test_case.assertEqual(obj1, obj2, msg=msg)


def assert_seed_allequal(a: Any, b: Any, err_msg: str = ""):
  """Backend-agnostic assertion to check if two seed objects are equal."""
  data_a = backend.get_seed_data(a)
  data_b = backend.get_seed_data(b)
  if data_a is None and data_b is None:
    return
  np.testing.assert_array_equal(data_a, data_b, err_msg=err_msg)


def assert_not_allequal(a: ArrayLike, b: ArrayLike, err_msg: str = ""):
  """Asserts that two objects are not element-wise equal."""
  np.testing.assert_(
      not np.array_equal(np.array(a), np.array(b)),
      msg=f"Arrays are unexpectedly equal.\n{err_msg}",
  )


def assert_seed_not_allequal(a: Any, b: Any, err_msg: str = ""):
  """Asserts that two seed objects are not element-wise equal."""
  data_a = backend.get_seed_data(a)
  data_b = backend.get_seed_data(b)
  if data_a is None and data_b is None:
    raise AssertionError(
        f"Seeds are unexpectedly equal (both are None). {err_msg}"
    )
  if data_a is None or data_b is None:
    return
  np.testing.assert_(
      not np.array_equal(data_a, data_b),
      msg=f"Seeds are unexpectedly equal.\n{err_msg}",
  )


def assert_all_finite(a: ArrayLike, err_msg: str = ""):
  """Backend-agnostic assertion to check if all elements in an array are finite.

  Args:
    a: The array-like object to check.
    err_msg: The error message to be printed in case of failure.

  Raises:
    AssertionError: If the array contains non-finite values.
  """
  if not np.all(np.isfinite(np.array(a))):
    raise AssertionError(err_msg or "Array contains non-finite values.")


def assert_all_non_negative(a: ArrayLike, err_msg: str = ""):
  """Backend-agnostic assertion to check if all elements are non-negative.

  Args:
    a: The array-like object to check.
    err_msg: The error message to be printed in case of failure.

  Raises:
    AssertionError: If the array contains negative values.
  """
  if not np.all(np.array(a) >= 0):
    raise AssertionError(err_msg or "Array contains negative values.")


# --- Proto Utilities ---
def normalize_tensor_protos(proto: message.Message):
  """Recursively normalizes TensorProto messages within a proto (In-place).

  This ensures a consistent serialization format across different backends
  (e.g., JAX vs TF) by repacking TensorProtos using the current backend's
  canonical method (backend.make_tensor_proto). This handles differences
  like using `bool_val` versus `tensor_content` for boolean tensors.

  Args:
    proto: The protobuf message object to normalize. This object is modified in
      place.
  """
  if not isinstance(proto, message.Message):
    return

  for desc, value in proto.ListFields():
    if desc.type != FieldDescriptor.TYPE_MESSAGE:
      continue

    # A map is defined as a repeated field whose message type has the
    # map_entry option set.
    is_map = (
        desc.label == FieldDescriptor.LABEL_REPEATED
        and desc.message_type.has_options
        and desc.message_type.GetOptions().map_entry
    )

    if is_map:
      for item in value.values():
        # Helper checks if values are scalars or messages.
        _process_message_for_normalization(item)

    elif desc.label == FieldDescriptor.LABEL_REPEATED:
      # Handle standard repeated message fields.
      for item in value:
        _process_message_for_normalization(item)
    else:
      # Handle singular message fields.
      _process_message_for_normalization(value)


def _process_message_for_normalization(msg: Any):
  """Helper to process a potential message during normalization traversal."""
  # Ensure we only process message objects.
  # If msg is a scalar (e.g., string from map<string, string>), stop recursion.
  if not isinstance(msg, message.Message):
    return

  if isinstance(msg, tensor_pb2.TensorProto):
    _repack_tensor_proto(msg)
  else:
    # If it's another message type, recurse into its fields.
    normalize_tensor_protos(msg)


def _repack_tensor_proto(tensor_proto: "tensor_pb2.TensorProto"):
  """Repacks a TensorProto in place to use a consistent serialization format."""
  if not tensor_proto.ByteSize():
    return

  try:
    data_array = backend.make_ndarray(tensor_proto)
  except Exception as e:
    raise ValueError(
        "Failed to deserialize TensorProto during normalization:"
        f" {e}\nProto content:\n{tensor_proto}"
    ) from e

  new_tensor_proto = backend.make_tensor_proto(data_array)

  tensor_proto.Clear()
  tensor_proto.CopyFrom(new_tensor_proto)


def assert_normalized_proto_equal(
    test_case: parameterized.TestCase,
    expected: message.Message,
    actual: message.Message,
    msg: Optional[str] = None,
    **kwargs: Any,
):
  """Compares two protos after normalizing TensorProto fields.

  Use this instead of compare.assertProtoEqual when protos contain tensors
  to ensure backend-agnostic comparison.

  Args:
    test_case: The TestCase instance (self).
    expected: The expected protobuf message.
    actual: The actual protobuf message.
    msg: An optional message to display on failure.
    **kwargs: Additional keyword arguments passed to assertProto2Equal (e.g.,
      precision).
  """
  # Work on copies to avoid mutating the original objects
  expected_copy = expected.__class__()
  expected_copy.CopyFrom(expected)
  actual_copy = actual.__class__()
  actual_copy.CopyFrom(actual)

  try:
    normalize_tensor_protos(expected_copy)
    normalize_tensor_protos(actual_copy)
  except ValueError as e:
    test_case.fail(f"Proto normalization failed: {e}. {msg}")

  compare.assertProtoEqual(
      test_case, expected_copy, actual_copy, msg=msg, **kwargs
  )


class MeridianTestCase(parameterized.TestCase):
  """Base test class for Meridian providing backend-aware utilities.

  This class handles initialization timing issues (crucial for JAX by forcing
  tensor operations into setUp) and provides a unified way to handle random
  number generation across backends (Stateful TF vs Stateless JAX).
  """

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    # Enforce determinism for TensorFlow tests before any tests are run.
    # This is a no-op with a warning for the JAX backend.
    backend.enable_op_determinism()

  def setUp(self):
    super().setUp()
    # Default seed, can be overridden by subclasses before calling
    # _initialize_rng().
    self.seed = 42
    self._jax_key = None
    self._initialize_rng()

  def _initialize_rng(self):
    """Initializes the RNG state or key based on self.seed."""
    current_backend = config.get_backend()

    if current_backend == config.Backend.TENSORFLOW:
      # In TF, we use the global stateful seed for test reproducibility.
      try:
        backend.set_random_seed(self.seed)
      except NotImplementedError:
        # Handle cases where backend might be misconfigured during transition.
        pass
    elif current_backend == config.Backend.JAX:
      # In JAX, we must manage PRNGKeys explicitly.
      # Import JAX locally to avoid hard dependency if TF is the active backend,
      # and to ensure initialization happens after absltest.main() starts.
      # pylint: disable=g-import-not-at-top
      import jax
      # pylint: enable=g-import-not-at-top
      self._jax_key = jax.random.PRNGKey(self.seed)
    else:
      raise ValueError(f"Unknown backend: {current_backend}")

  def get_next_rng_seed_or_key(self) -> Optional[Any]:
    """Gets the next available seed or key for backend operations.

    This should be passed to the `seed` argument of TFP sampling methods.

    Returns:
      A JAX PRNGKey if the backend is JAX (splitting the internal key).
      None if the backend is TensorFlow (relying on the global state).
    """
    if self._jax_key is not None:
      # JAX requires splitting the key for each use.
      # pylint: disable=g-import-not-at-top
      import jax
      # pylint: enable=g-import-not-at-top
      self._jax_key, subkey = jax.random.split(self._jax_key)
      return subkey
    else:
      # For stateful TF, returning None allows TFP/TF to use the global seed.
      return None

  def sample(
      self,
      distribution: backend.tfd.Distribution,
      sample_shape: Any = (),
      **kwargs: Any,
  ) -> backend.Tensor:
    """Performs a backend-agnostic sample from a distribution.

    This method abstracts away the need for explicit seed management in JAX.
    When the JAX backend is active, it automatically provides a PRNGKey from
    the test's managed key state. In TensorFlow, it performs a standard sample.

    Args:
      distribution: The TFP distribution object to sample from.
      sample_shape: The shape of the desired sample.
      **kwargs: Additional keyword arguments to pass to the underlying `sample`
        method (e.g., `name`).

    Returns:
      A tensor containing the sampled values.
    """
    seed = self.get_next_rng_seed_or_key()
    return distribution.sample(sample_shape=sample_shape, seed=seed, **kwargs)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides tools for converting and wrapping MMM schema data.

This package contains modules to transform Marketing Mix Modeling (MMM) protocol
buffer data into other formats and provides high-level wrappers for easier data
manipulation, analysis, and reporting.
"""

from scenarioplanner.converters import dataframe
from scenarioplanner.converters import mmm
from scenarioplanner.converters import mmm_converter
from scenarioplanner.converters import sheets
